#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
import shutil
import math


def _string_with_square(arr, sep):
    return f"[{sep.join(map(str, arr))}]"


def _find_unit_range(arr):
    if len(arr) < 2:
        return arr

    narr = []
    rangestart = arr[0]
    rangestop = arr[0]

    for v in arr[1:]:
        if v == rangestop + 1:
            rangestop = v
        elif rangestart == rangestop:
            narr.append(rangestart)
            rangestart = v
            rangestop = v
        else:
            narr.append(range(rangestart, rangestop + 1))
            rangestart = v
            rangestop = v

    if rangestart == rangestop:
        narr.append(rangestart)
    else:
        narr.append(range(rangestart, rangestop + 1))

    return narr


def _gate_name_padding(qubits, bits, zvars):
    nq = len(qubits)
    qubits_padding = 0 if nq in [1, 0] else math.floor(math.log10(nq)) + 2

    bitspadding = 0 if not bits else len(str(bits)) + 1
    zvarspadding = 0 if not zvars else len(str(zvars)) + 1

    return max(qubits_padding, bitspadding, zvarspadding)


class AsciiCanvas:
    def __init__(self, width=None):
        if width is None:
            _, self.width = 0, shutil.get_terminal_size().columns
        self.width = width
        self.data = []

    def get_rows(self):
        return len(self.data)

    def get_cols(self):
        return self.width

    def push_line(self):
        self.data.append([" "] * self.width)

    def __getitem__(self, position):
        row, col = position
        if row >= len(self.data):
            return " "
        return self.data[row][col]

    def __setitem__(self, position, value):
        row, col = position
        while row >= self.get_rows():
            self.push_line()
        while col >= len(self.data[row]):
            self.data[row].append(" ")
        self.data[row][col] = value

    def __str__(self):
        return "\n".join("".join(row) for row in self.data if row)

    def draw_hline(self, row, col, width):
        start_col, stop_col = min(col, col + width - 1), max(col, col + width - 1)

        for i in range(start_col, stop_col + 1):
            current_char = self[row, i]
            # Check and replace characters based on conditions
            if current_char == "│":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "├", "┼", "┤"
                )
            elif current_char == "┤":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "┼", "┼", "┤"
                )
            elif current_char == "├":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "├", "┼", "┼"
                )
            elif current_char == "╵":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "└", "┴", "┘"
                )
            elif current_char == "└":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "└", "┴", "┴"
                )
            elif current_char == "┘":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "┴", "┴", "┘"
                )
            elif current_char == "╷":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "┌", "┬", "┐"
                )
            elif current_char == "║":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╟", "╫", "╢"
                )
            elif current_char == "╟":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╟", "╫", "╫"
                )
            elif current_char == "╢":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╫", "╫", "╢"
                )
            elif current_char == "╶":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╶", "─", "─"
                )
            elif current_char == "╴":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "─", "─", "╴"
                )
            else:
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╶", "─", "╴"
                )

        return self

    def _start_mid_stop(self, current, start, stop, start_char, mid_char, stop_char):
        if current == start:
            return start_char
        elif current == stop:
            return stop_char
        return mid_char

    def draw_vline(self, row, col, height):
        start_row, stop_row = min(row, row + height - 1), max(row, row + height - 1)
        for i in range(start_row, stop_row + 1):
            current_char = self[i, col]
            if current_char == "─":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┬", "┼", "┴"
                )
            elif current_char == "┴":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┼", "┼", "┴"
                )
            elif current_char == "┬":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┬", "┼", "┼"
                )
            elif current_char == "╴":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┐", "┤", "┘"
                )
            elif current_char == "┐":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┐", "┤", "┤"
                )
            elif current_char == "┘":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┤", "┤", "┘"
                )
            elif current_char == "╶":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┌", "├", "└"
                )
            elif current_char == "┌":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "┌", "├", "├"
                )
            elif current_char == "└":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "├", "├", "└"
                )
            elif current_char == "╵":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "│", "│", "╵"
                )
            elif current_char == "╷":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╷", "│", "│"
                )
            elif current_char == "═":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╤", "╪", "╧"
                )
            elif current_char == "╤":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╤", "╪", "╪"
                )
            elif current_char == "╧":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╪", "╪", "╧"
                )
            else:
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╷", "│", "╵"
                )

        return self

    def draw_double_hline(self, row, col, width):
        start_col, stop_col = min(col, col + width - 1), max(col, col + width - 1)
        for i in range(start_col, stop_col + 1):
            if self[row, i] == "│":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╞", "╪", "╡"
                )
            elif self[row, i] == "╞":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╞", "╪", "╪"
                )
            elif self[row, i] == "╡":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╪", "╪", "╡"
                )
            elif self[row, i] == "╵":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╘", "╧", "╛"
                )
            elif self[row, i] == "╘":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╘", "╧", "╧"
                )
            elif self[row, i] == "╛":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╧", "╧", "╛"
                )
            elif self[row, i] == "╷":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╒", "╤", "╕"
                )
            elif self[row, i] == "╒":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╒", "╤", "╤"
                )
            elif self[row, i] == "╕":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╤", "╤", "╕"
                )
            elif self[row, i] == "║":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╠", "╬", "╣"
                )
            elif self[row, i] == "╠":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╠", "╬", "╣"
                )
            elif self[row, i] == "╣":
                self[row, i] = self._start_mid_stop(
                    i, start_col, stop_col, "╠", "╬", "╣"
                )
            else:
                self[row, i] = "═"

    def draw_double_vline(self, row, col, height):
        start_row, stop_row = min(row, row + height - 1), max(row, row + height - 1)
        for i in range(start_row, stop_row + 1):
            if self[i, col] == "─":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╥", "╫", "╨"
                )
            elif self[i, col] == "╥":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╥", "╫", "╫"
                )
            elif self[i, col] == "╨":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╫", "╫", "╨"
                )
            elif self[i, col] == "╶":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╓", "╟", "╙"
                )
            elif self[i, col] == "╓":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╓", "╟", "╟"
                )
            elif self[i, col] == "╙":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╟", "╟", "╙"
                )
            elif self[i, col] == "╴":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╖", "╢", "╜"
                )
            elif self[i, col] == "╖":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╖", "╢", "╢"
                )
            elif self[i, col] == "╜":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╢", "╢", "╜"
                )
            elif self[i, col] == "═":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╦", "╬", "╩"
                )
            elif self[i, col] == "╩":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╬", "╬", "╩"
                )
            elif self[i, col] == "╦":
                self[i, col] = self._start_mid_stop(
                    i, start_row, stop_row, "╦", "╬", "╬"
                )
            else:
                self[i, col] = "║"

    def draw_fill(self, char, row, col, width, height):
        for i in range(row, row + height):
            for j in range(col, col + width):
                self[i, j] = char

    def draw_empty(self, row, col, width, height):
        for i in range(row, row + height):
            while i >= len(self.data):
                self.push_line()
            for j in range(col, col + width):
                while j >= len(self.data[i]):
                    self.data[i].append(" ")
                if j == col and self.data[i][j] == "─":
                    self.data[i][j] = "╴"
                elif j == col + width - 1 and self.data[i][j] == "─":
                    self.data[i][j] = "╶"
                elif i == row and self.data[i][j] == "│":
                    self.data[i][j] = "╵"
                elif i == row + height - 1 and self.data[i][j] == "│":
                    self.data[i][j] = "╷"
                else:
                    self.data[i][j] = " "

    def draw_box(self, row, col, width, height, clean=False):
        if clean:
            self.draw_empty(row, col, width, height)

        self.draw_hline(row, col, width)
        self.draw_hline(row + height - 1, col, width)
        self.draw_vline(row, col, height)
        self.draw_vline(row, col + width - 1, height)
        return self

    def draw_text(self, text, row, col):
        for i, char in enumerate(text):
            self[row, col + i] = char
        return self

    def draw_vtext(self, text, row, col):
        for i, char in enumerate(text):
            self[row + i, col] = char
        return self

    def reset(self):
        self.data = []
        return self


class AsciiCircuit:
    def __init__(self, width=None):
        if width is None:
            _, width = 0, shutil.get_terminal_size().columns
        self.canvas = AsciiCanvas(width)
        self.qubitrow = {}
        self.bitrow = None
        self.zvarrow = None
        self.nonerow = None
        self.currentcol = 0

    def set_current_col(self, col):
        self.currentcol = max(self.currentcol, col)

    def get_current_col(self):
        return self.currentcol

    def get_qubit_row(self, qubit):
        return self.qubitrow.get(qubit)

    def get_bit_row(self):
        return self.bitrow

    def get_zvars_row(self):
        return self.zvarrow

    def reset(self):
        self.canvas.reset()
        self.qubitrow = {}
        self.bitrow = None
        self.zvars = None
        self.nonerow = None
        self.currentcol = 1
        return self

    def draw_wires(self, qubits, bits, zvars):
        # Draw qubit wires and labels
        for i, q in enumerate(qubits):
            row = i * 2 + 1
            qubitstr = f"q[{q}]: "
            self.canvas.draw_text(qubitstr, row, 1)
            self.qubitrow[q] = row
            self.set_current_col(len(qubitstr) + 1)

        if len(bits) > 0:
            row = len(qubits) * 2 + 3
            bitstr = "c: "
            self.canvas.draw_text(bitstr, row, 1)
            self.bitrow = row
            self.set_current_col(len(bitstr) + 1)

        if len(zvars) > 0:
            if len(bits) < 1:
                row = (len(qubits)) * 2 + 3
            else:
                row = (len(qubits) + 1) * 2 + 3

            zstr = "z: "
            self.canvas.draw_text(zstr, row, 1)
            self.zvarrow = row
            self.set_current_col(len(zstr) + 1)

        ccol = self.get_current_col() - 3

        ccol = self.get_current_col()

        # Fill space and draw horizontal line for qubits
        for i in range(len(qubits)):
            row = i * 2
            self.canvas.draw_fill(" ", row, ccol, self.canvas.get_cols() - ccol, 1)
            self.canvas.draw_hline(row + 1, ccol, self.canvas.get_cols() - ccol)

            self.canvas.draw_fill(
                " ", (len(qubits) - 1) * 2 + 3, ccol, self.canvas.get_cols() - ccol, 1
            )

        # Prepare space and draw double horizontal line for bits if more than one bit
        if len(bits) > 0:
            row = len(qubits) * 2 + 2
            self.canvas.draw_fill(" ", row, ccol, self.canvas.get_cols() - ccol, 1)
            self.canvas.draw_double_hline(row + 1, ccol, self.canvas.get_cols() - ccol)

        if len(zvars) > 0:
            if len(bits) < 1:
                row = (len(qubits)) * 2 + 2
            else:
                row = (len(qubits) + 1) * 2 + 2

            self.canvas.draw_fill(" ", row, ccol, self.canvas.get_cols() - ccol, 1)
            self.canvas.draw_double_hline(row + 1, ccol, self.canvas.get_cols() - ccol)

        self.set_current_col(ccol + 1)
        return self

    def draw_operation(self, operation, qubits, bits=None, zvars=None):
        # Initialize default values for bits and zvars if they are None
        bits = bits or []
        zvars = zvars or []

        # Calculate padding for qubits, bits, and zvars
        namepadding = _gate_name_padding(qubits, bits, zvars)
        ccol = self.get_current_col()

        qubitrow = [self.get_qubit_row(q) for q in qubits]
        bitrow = self.get_bit_row()
        zvarrow = self.get_zvars_row()

        if not qubits and not bits and not zvars:
            # Draw the None row only if it's not already drawn
            if self.nonerow is None:
                none_row_start = (
                    len(self.qubitrow)
                    + (1 if self.bitrow else 0)
                    + (1 if self.zvarrow else 0)
                ) * 2 + 3
                self.nonerow = none_row_start
                ccol = 0
                self.canvas.draw_fill(
                    " ", none_row_start, ccol, self.canvas.get_cols(), 1
                )
                self.canvas.draw_double_hline(
                    none_row_start, ccol, self.canvas.get_cols()
                )

            # Now draw the operation on the None line
            none_row = self.nonerow
            ccol = self.get_current_col()
            gw = operation.asciiwidth([], [], [])
            self.canvas.draw_box(none_row - 1, ccol, gw, 3, clean=True)  # 3 rows height
            self.canvas.draw_text(str(operation), none_row, ccol + 1)
            self.set_current_col(ccol + gw)
            return self

        # Determine start and stop rows for qubits, bits, and zvars
        if bits or zvars:
            qubit_startrow = min(qubitrow, default=float("inf"))
            qubit_stoprow = max(qubitrow, default=0)

            bits_startrow = bitrow if bits else float("inf")
            bits_stoprow = bitrow if bits else 0

            zvar_startrow = zvarrow if zvars else float("inf")
            zvar_stoprow = zvarrow if zvars else 0

            # Handle case where both qubits and zvars are present
            if qubitrow and zvars:
                # Draw box for qubits
                qubit_gateheight = qubit_stoprow - qubit_startrow + 1
                qubit_midrow = qubit_startrow + qubit_gateheight // 2
                gw = operation.asciiwidth(qubits, bits, [])
                # Calculate zvar width first
                gw_zvar = operation.asciiwidth([], [], zvars)

                # Ensure both boxes have the same width
                gw = max(gw, gw_zvar)
                midcol = ccol + gw // 2

                self.canvas.draw_box(
                    qubit_startrow - 1, ccol, gw, qubit_gateheight + 2, clean=True
                )
                self.canvas.draw_text(
                    str(operation), qubit_midrow, ccol + namepadding - 1
                )

                # Draw zvars without a box, with adjusted spacing
                self.canvas.draw_text(",".join(map(str, zvars)), zvarrow + 1, midcol)

                # Draw vertical double line connecting qubits and zvars

                self.canvas.draw_double_vline(
                    qubit_stoprow + 1, midcol, zvar_startrow - qubit_stoprow
                )

                endcol = ccol + gw

            elif qubitrow and bits:
                # Draw box for qubits
                qubit_gateheight = qubit_stoprow - qubit_startrow + 1
                qubit_midrow = qubit_startrow + qubit_gateheight // 2
                gw_qubits = operation.asciiwidth(qubits, [], [])
                gw_bits = operation.asciiwidth([], bits, [])

                # Ensure the box for qubits is as wide as needed for both qubits and bits
                gw = max(gw_qubits, gw_bits)
                midcol = ccol + gw // 2

                self.canvas.draw_box(
                    qubit_startrow - 1, ccol, gw, qubit_gateheight + 2, clean=True
                )
                self.canvas.draw_text(
                    str(operation), qubit_midrow, ccol + namepadding - 1
                )

                # Draw the bits below the qubits, centered horizontally
                bitstr = ",".join(map(str, bits))
                bit_start_col = midcol - len(bitstr) // 2
                self.canvas.draw_text(bitstr, bits_startrow + 1, bit_start_col)

                # Draw vertical double line connecting qubits and bits
                self.canvas.draw_double_vline(
                    qubit_stoprow + 1, midcol, bits_stoprow - qubit_stoprow
                )

                endcol = ccol + gw

            elif zvars and not qubitrow:
                startrow = min(qubit_startrow, zvar_startrow) - 1
                stoprow = max(qubit_stoprow, zvar_stoprow) + 1
                gateheight = stoprow - startrow + 1
                midrow = startrow + gateheight // 2

                # Draw a single box
                gw = operation.asciiwidth(qubits, bits, zvars)
                self.canvas.draw_box(startrow, ccol, gw, gateheight, clean=True)
                zstr = f"{','.join(map(str, zvars))} ═> "
                self.canvas.draw_text(zstr, midrow, ccol + 1)
                text_start_col = ccol + len(zstr) + 1
                # Draw the operation name
                self.canvas.draw_text(str(operation), midrow, text_start_col)

                midcol = ccol + gw // 2
                endcol = ccol + gw

            elif bits and not qubitrow:
                startrow = min(qubit_startrow, bits_startrow) - 1
                stoprow = max(qubit_stoprow, bits_stoprow) + 1
                gateheight = stoprow - startrow + 1
                midrow = startrow + gateheight // 2

                # Draw a single box
                gw = operation.asciiwidth(qubits, bits, zvars)
                self.canvas.draw_box(startrow, ccol, gw, gateheight, clean=True)
                zstr = f"{','.join(map(str, bits))} ═> "
                self.canvas.draw_text(zstr, midrow, ccol + 1)
                text_start_col = ccol + len(zstr) + 1
                # Draw the operation name
                self.canvas.draw_text(str(operation), midrow, text_start_col)

                midcol = ccol + gw // 2
                endcol = ccol + gw

        else:
            startrow = min(qubitrow) - 1
            stoprow = max(qubitrow) + 1
            gateheight = stoprow - startrow + 1
            midrow = startrow + gateheight // 2

            # Draw the box around the operation
            gw = operation.asciiwidth(qubits, bits, zvars)
            self.canvas.draw_box(startrow, ccol, gw, gateheight, clean=True)
            # Properly centered
            self.canvas.draw_text(str(operation), midrow, ccol + namepadding + 1)
            midcol = ccol + gw // 2
            endcol = ccol + gw + namepadding

        # If more than one qubit, label each qubit within the box
        if len(qubits) > 1:
            for i, qr in enumerate(qubitrow):
                self.canvas.draw_text(str(i), qr, ccol + 1)
        self.set_current_col(endcol)

        return self

    def draw_control(self, operation, qubits, _):
        if not isinstance(operation, mc.Control):
            raise TypeError("operation must be an instance of Operation")

        if operation.op.num_qubits == 1:
            target_row = self.get_qubit_row(qubits[-1])
            control_rows = [self.get_qubit_row(q) for q in qubits[:-1]]
            max_row = max(control_rows + [target_row])
            min_row = min(control_rows + [target_row])
            current_col = self.get_current_col()
            gate_width = operation.get_operation().asciiwidth([qubits[-1]], [], [])
            middle_column = current_col + gate_width // 2

            self.canvas.draw_vline(min_row, middle_column, max_row - min_row + 1)

            for row in control_rows:
                self.canvas[row, middle_column] = "●"

            self.draw_operation(operation.get_operation(), [qubits[-1]], [], [])
        else:
            self.draw_operation(operation, qubits, [])

    def draw_barrier(self, barrier, qubits, bits, zvars):
        for qubit in qubits:
            qubit_row = self.get_qubit_row(qubit)
            current_col = self.get_current_col()

            self.canvas.draw_text("░", qubit_row, current_col)
            self.canvas.draw_text("░", qubit_row, current_col)
            self.canvas.draw_text("░", qubit_row + 1, current_col)

        self.set_current_col(current_col + 1)
        return self

    def draw_ifstatement(self, g, qubits, bits, zvars):
        qubitrow = [self.get_qubit_row(q) for q in qubits]
        bitrow = self.get_bit_row()

        nb = len(bits)
        val = g.get_bitstring()

        ccol = self.get_current_col()

        bstr = _string_with_square(_find_unit_range(bits), ",")
        btext = f"c{bstr}==" + val.to01()

        self.canvas.draw_box(bitrow - 1, ccol, len(btext) + 2, 3, clean=True)
        self.canvas.draw_text(btext, bitrow, ccol + 1)

        self.set_current_col(ccol + len(btext) + 2)
        ifcol = self.get_current_col()

        qstartrow = min(qubitrow) - 1
        qstoprow = max(qubitrow) + 1
        qw = g.op.asciiwidth(qubits, [], [])
        qh = qstoprow - qstartrow + 1
        qmh = qstartrow + qh // 2
        namepadding = _gate_name_padding(qubits, [], [])

        self.canvas.draw_box(qstartrow, ifcol, qw, qh, clean=True)

        for i, qr in enumerate(qubitrow):
            self.canvas.draw_text(str(i), qr, ifcol + 1)

        self.canvas.draw_text(repr(g.op), qmh, ifcol + namepadding + 1)

        midcol = ifcol + qw // 2

        self.canvas.draw_double_vline(qstoprow, midcol, bitrow - qstoprow)
        self.canvas.draw_double_hline(bitrow - 1, ifcol, midcol - ifcol)
        self.canvas.draw_text("╝", bitrow - 1, midcol)
        self.canvas.draw_text("○", bitrow - 1, ifcol)

        self.set_current_col(ifcol + qw)

    def draw_reset(self, reset, qubits, _, zvars):
        if not isinstance(reset, mc.Reset):
            raise TypeError("Must be a Reset Operation")

        qubit_row = self.get_qubit_row(qubits[0])
        current_col = self.get_current_col()
        self.canvas.draw_box(qubit_row - 1, current_col, 5, 3, clean=True)

        # Draw the reset symbol (|0⟩) in the center of the box
        self.canvas.draw_text("|0⟩", qubit_row, current_col + 1)

        # Move the current column index forward to prevent overlapping with future elements
        self.set_current_col(current_col + 5)
        return self

    def draw_parallel(self, parallel, qubits, _, zvars):
        if not isinstance(parallel, mc.Parallel):
            raise TypeError("Must be a Parallel")

        op = parallel.get_operation()
        nq = op.num_qubits
        nb = op.num_bits
        ccol = self.get_current_col()

        for i in range(parallel.num_repeats):
            self.currentcol = ccol
            self.draw_operation(op, qubits[nq * i : nq * (i + 1)], [], [])

        return self

    def draw_paulistring(self, g, qubits, bits, zvars):
        ccol = self.get_current_col()
        qubitrow = [self.get_qubit_row(q) for q in qubits]

        startrow = min(qubitrow) - 1
        stoprow = max(qubitrow) + 1
        gateheight = stoprow - startrow + 1

        gw = g.asciiwidth(qubits, bits, zvars)

        self.canvas.draw_box(startrow, ccol, gw, gateheight, clean=True)

        for i, qr in enumerate(qubitrow):
            pauli_char = g.pauli[i]
            label = f"{i}: {pauli_char}" if len(qubits) > 1 else f"{pauli_char}"
            self.canvas.draw_text(label, qr, ccol + 1)

        self.set_current_col(ccol + gw)

        return self

    def draw_instruction(self, instruction):
        if not isinstance(instruction, mc.Instruction):
            raise TypeError("Must be an Instruction")

        return self.draw_operation(
            instruction.get_operation(),
            instruction._qubits,
            instruction._bits,
            instruction._zvars,
        )


__all__ = ["AsciiCircuit", "AsciiCanvas"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
from mimiqcircuits.operations.utils import power_idempotent
from mimiqcircuits.operations.utils import control_one_defined
from symengine import pi, sqrt, Matrix, I


class GateH(mc.Gate):
    r"""Single qubit Hadamard gate.

    **Matrix representation:**

    .. math::
        \operatorname{H} = \frac{1}{\sqrt{2}} \begin{pmatrix}
            1 & 1 \\
            1 & -1
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateH()
        H
        >>> GateH().matrix()
        [0.707106781186548, 0.707106781186548]
        [0.707106781186548, -0.707106781186548]
        <BLANKLINE>
        >>> c = Circuit().push(GateH(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── H @ q[0]
        <BLANKLINE>
        >>> GateH().power(2), GateH().inverse()
        (ID, H)
        >>> GateH().decompose()
        1-qubit circuit with 1 instructions:
        └── U((1/2)*pi, 0, pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "H"

    _num_qubits = 1
    _qregsizes = [1]

    def inverse(self):
        return self

    def _power(self, p):
        # H^(2n) = ID
        # H^(2n+1) = H
        return power_idempotent(self, p)

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCH(), mc.Control(2, GateH()))

    def _matrix(self):
        return Matrix([[1, 1], [1, -1]]) / sqrt(2)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateU(pi / 2, 0, pi), q)
        return circ


class GateHXY(mc.Gate):
    r"""Single qubit HXY gate.

    **Matrix representation:**

    .. math::
        \operatorname{HXY} = \frac{1}{\sqrt{2}} \begin{pmatrix}
            0 & 1 - i \\
            1 + i & 0
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateHXY()
        HXY
        >>> GateHXY().matrix()
        [0, 0.707106781186548 - 0.707106781186548*I]
        [0.707106781186548 + 0.707106781186548*I, 0]
        <BLANKLINE>
        >>> c = Circuit().push(GateHXY(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── HXY @ q[0]
        <BLANKLINE>
        >>> GateHXY().power(2), GateHXY().inverse()
        (HXY**2, HXY)
        >>> GateHXY().decompose()
        1-qubit circuit with 5 instructions:
        ├── H @ q[0]
        ├── Z @ q[0]
        ├── H @ q[0]
        ├── S @ q[0]
        └── U(0, 0, 0, (-1/4)*pi) @ q[0]
        <BLANKLINE>
    """
    _name = "HXY"

    _num_qubits = 1
    _qregsizes = [1]

    def inverse(self):
        return self

    def _power(self, p):
        return mc.Power(GateHXY(), p)

    def _control(self, n):
        return control_one_defined(n, self, mc.Control(1, GateHXY()))

    def _matrix(self):
        return Matrix([[0, 1-I], [1+I, 0]]) / sqrt(2)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateH(), q)
        circ.push(mc.GateZ(), q)
        circ.push(mc.GateH(), q)
        circ.push(mc.GateS(), q)
        circ.push(mc.GateU(0, 0, 0, -pi/4), q)
        return circ


class GateHYZ(mc.Gate):
    r"""Single qubit HYZ gate.

    **Matrix representation:**

    .. math::
        \operatorname{HYZ} = \frac{1}{\sqrt{2}} \begin{pmatrix}
            1 & -i \\
            i & -1
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateHYZ()
        HYZ
        >>> GateHYZ().matrix()
        [0.707106781186548, -0.0 - 0.707106781186548*I]
        [0.0 + 0.707106781186548*I, -0.707106781186548]
        <BLANKLINE>
        >>> c = Circuit().push(GateHYZ(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── HYZ @ q[0]
        <BLANKLINE>
        >>> GateHYZ().power(2), GateHYZ().inverse()
        (HYZ**2, HYZ)
        >>> GateHYZ().decompose()
        1-qubit circuit with 5 instructions:
        ├── H @ q[0]
        ├── S @ q[0]
        ├── H @ q[0]
        ├── Z @ q[0]
        └── U(0, 0, 0, (-1/4)*pi) @ q[0]
        <BLANKLINE>
    """
    _name = "HYZ"

    _num_qubits = 1
    _qregsizes = [1]

    def inverse(self):
        return self

    def _power(self, p):
        return mc.Power(GateHYZ(), p)

    def _control(self, n):
        return control_one_defined(
            n, self, mc.Control(1, GateHYZ()), mc.Control(2, GateHYZ())
        )

    def _matrix(self):
        return Matrix([[1, -I], [I, -1]]) / sqrt(2)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateH(), q)
        circ.push(mc.GateS(), q)
        circ.push(mc.GateH(), q)
        circ.push(mc.GateZ(), q)
        circ.push(mc.GateU(0, 0, 0, -pi/4), q)
        return circ


class GateHXZ(mc.Gate):
    r"""Single qubit HXZ gate (alias for :class:`GateH`).

    **Matrix representation:**

    .. math::
        \operatorname{H} = \frac{1}{\sqrt{2}} \begin{pmatrix}
            1 & 1 \\
            1 & -1
        \end{pmatrix}

    Examples:

        The HXZ gate behaves exactly like the Hadamard gate:

        >>> from mimiqcircuits import *
        >>> GateHXZ()
        H
        >>> GateHXZ().matrix()
        [0.707106781186548, 0.707106781186548]
        [0.707106781186548, -0.707106781186548]
        <BLANKLINE>

        Adding GateHXZ to a circuit:

        >>> c = Circuit().push(GateHXZ(), 0)

        Power and inverse of the gate:

        >>> GateHXZ().power(2), GateHXZ().inverse()
        (ID, H)

        Decomposition of the gate:

        >>> GateHXZ().decompose()
        1-qubit circuit with 1 instructions:
        └── U((1/2)*pi, 0, pi, 0.0) @ q[0]
        <BLANKLINE>
    """
    def __new__(self):
        return GateH()



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits import Gate, GateID, Circuit


class Delay(Gate):
    r"""
    Delay(t)

    1-qubit delay gate

    This gate is equivalent to a GateID gate, except that it is parametrized by
    a time parameter `t`. The parameter does not affect the action of the gate. The only
    purpose of this gate is to act as a placeholder to indicate idle noise, in
    which case the parameter `t` can later be used to further specify the noise properties.

    The gate can be created by calling `Delay(t)` where `t` is a number.

    **Matrix representation:**

    .. math::
        \operatorname{Delay}(t) =
        \begin{pmatrix}
            1 & 0 \\
            0 & 1
        \end{pmatrix}


    Examples:
        >>> Delay(0.3)
        Delay(0.3)

        >>> delay_gate = Delay(0.1)
        >>> delay_gate.matrix()
        [1.0, 0]
        [0, 1.0]
        <BLANKLINE>

        >>> c = Circuit()
        >>> c.push(delay_gate, 0)
        1-qubit circuit with 1 instructions:
        └── Delay(0.1) @ q[0]
        <BLANKLINE>

        Decomposition

        >>> delay_gate = Delay(0.2)
        >>> delay_gate.decompose()
        1-qubit circuit with 1 instructions:
        └── ID @ q[0]
        <BLANKLINE>
    """

    _name = "Delay"
    _num_qubits = 1
    _num_qregs = 1
    _num_cregs = 0
    _num_bits = 0
    _parnames = "t"

    def __init__(self, t: float = 1):
        self.t = t
        super().__init__()

    def __str__(self):
        return f"Delay({self.t})"

    def _matrix(self):
        """Return the matrix representation of the Delay gate."""
        return GateID().matrix()

    def inverse(self):
        raise ValueError("Cannot invert a Delay gate.")

    def power(self, _):
        raise ValueError("Cannot raise a Delay gate to any power.")

    def _decompose(self, circ: Circuit, qubits, bitsm, zvars):
        """Decompose the Delay gate into a GateID for the given circuit."""
        for qubit in qubits:
            circ.push(GateID(), [qubit])
        return circ


__all__ = ["Delay"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc


class GateCH(mc.Control):
    r"""Two qubit Controlled-Hadamard gate.

    By convention, the first qubit is the control and the second is
    the target

    **Matrix representation:**

    .. math::
        \operatorname{CH} = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & \frac{1}{\sqrt{2}}  & \frac{1}{\sqrt{2}}  \\
            0 & 0 & \frac{1}{\sqrt{2}}  & -\frac{1}{\sqrt{2}}
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCH(), GateCH().num_controls, GateCH().num_targets, GateCH().num_qubits
        (CH, 1, 1, 2)
        >>> GateCH().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 0.707106781186548, 0.707106781186548]
        [0, 0, 0.707106781186548, -0.707106781186548]
        <BLANKLINE>
        >>> c = Circuit().push(GateCH(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CH @ q[0], q[1]
        <BLANKLINE>
        >>> GateCH().power(2), GateCH().inverse()
        (CID, CH)
        >>> GateCH().decompose()
        2-qubit circuit with 7 instructions:
        ├── S @ q[1]
        ├── H @ q[1]
        ├── T @ q[1]
        ├── CX @ q[0], q[1]
        ├── T† @ q[1]
        ├── H @ q[1]
        └── S† @ q[1]
        <BLANKLINE>

    """

    def __init__(self):
        super().__init__(1, mc.GateH())

    def _decompose(self, circ, qubits, bits, zvars):
        c, t = qubits

        circ.push(mc.GateS(), t)
        circ.push(mc.GateH(), t)
        circ.push(mc.GateT(), t)
        circ.push(mc.GateCX(), c, t)
        circ.push(mc.GateTDG(), t)
        circ.push(mc.GateH(), t)
        circ.push(mc.GateSDG(), t)

        return circ



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from typing import List, Union
import symengine as se
import mimiqcircuits as mc
import sympy as sp
from mimiqcircuits.operations.krauschannel import krauschannel
from mimiqcircuits.operations.rescaledgates import RescaledGate
from mimiqcircuits.operations.gates.custom import GateCustom
from mimiqcircuits.operations.gates.gate import Gate


def _is_valid_power_of_2(n):
    return n > 0 and (n & (n - 1)) == 0


def _convert_to_symengine(matrix):
    if isinstance(matrix, np.ndarray):
        return se.Matrix(matrix.tolist())
    elif isinstance(matrix, sp.Matrix):
        return se.Matrix(matrix.tolist())
    elif isinstance(matrix, se.Matrix):
        return matrix
    else:
        raise ValueError(f"Unsupported matrix type: {type(matrix)}")


class MixedUnitary(krauschannel):
    r"""MixedUnitary(p,umatrices).

    Custom N-qubit mixed unitary channel specified by a list of :math:`2^N \times 2^N`
    unitary matrices and a list of probabilities that add up to 1.

    A mixed unitary noise channel is defined by:

    .. math::
        \mathcal{E}(\rho) = \sum_k p_k U_k \rho U_k^\dagger,

    where :math:`0 \leq p_k \leq 1` and :math:`U_k` are unitary matrices.
    The probabilities must fulfill :math:`\sum_k p_k = 1`.

    If your Kraus matrices are not all proportional to unitaries, use :func:`KrausChannel` instead.

    The unitary matrices are defined in the computational basis in the usual textbook order
    (the first qubit corresponds to the left-most qubit).
    For 1 qubit, we have :math:`|0\rangle`, :math:`|1\rangle`.
    For 2 qubits, we have :math:`|00\rangle`, :math:`|01\rangle`, :math:`|10\rangle`, :math:`|11\rangle`.

    **Note:** Currently, only 1 and 2-qubit custom MixedUnitary channels are supported.

    See Also:
        :class:`Kraus`
        :func:`ismixedunitary`
        :class:`krauschannel`

    Parameters:
        p (list): List of probabilities, must be positive real numbers and add up to 1.
        umatrices (list): List of complex-valued :math:`2^N \times 2^N` matrices. The number of qubits is equal to :math:`N`.

    The length of the lists `p` and `umatrices` must be equal.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> c = Circuit()
        >>> c.push(MixedUnitary([0.9, 0.1], [Matrix([[1, 0], [0, 1]]), Matrix([[0, 1], [1, 0]])]), 0)
        1-qubit circuit with 1 instructions:
        └── MixedUnitary((0.9, "Custom([1 0; 0 1])"),(0.1, "Custom([0 1; 1 0])")) @ q[0]
        <BLANKLINE>

        >>> c.push(MixedUnitary([0.8, 0.2], [Matrix(GateID().matrix()), Matrix(GateRX(0.2).matrix())]), 1)
        2-qubit circuit with 2 instructions:
        ├── MixedUnitary((0.9, "Custom([1 0; 0 1])"),(0.1, "Custom([0 1; 1 0])")) @ q[0]
        └── MixedUnitary((0.8, "Custom([1.0 0; 0 1.0])"),(0.2, "Custom([0.995004165278026 -0.0 - 0.0998334166468282*I; -0.0 - 0.0998334166468282*I 0.995004165278026])")) @ q[1]
        <BLANKLINE>
    
        RescaleGate

        >>> p1= 0.2
        >>> p2 = 0.8
        >>> U1 = Matrix([[1, 0], [0, 1]])  # Identity matrix
        >>> U2 = Matrix([[0, 1], [1, 0]])  # Pauli-X matrix
        >>> gate1 = GateCustom(U1)
        >>> gate2 = GateCustom(U2)
        >>> rescaled_gate1 = RescaledGate(gate1, sqrt(p1))
        >>> rescaled_gate2 = RescaledGate(gate2, sqrt(p2))
        >>> op = MixedUnitary([rescaled_gate1, rescaled_gate2])
        >>> op
        MixedUnitary((0.2, "Custom([1 0; 0 1])"), (0.8, "Custom([0 1; 1 0])"))
        >>> c.push(op,1)
        2-qubit circuit with 3 instructions:
        ├── MixedUnitary((0.9, "Custom([1 0; 0 1])"),(0.1, "Custom([0 1; 1 0])")) @ q[0]
        ├── MixedUnitary((0.8, "Custom([1.0 0; 0 1.0])"),(0.2, "Custom([0.995004165278026 -0.0 - 0.0998334166468282*I; -0.0 - 0.0998334166468282*I 0.995004165278026])")) @ q[1]
        └── MixedUnitary((0.2, "Custom([1 0; 0 1])"),(0.8, "Custom([0 1; 1 0])")) @ q[1]
        <BLANKLINE>
    """

    _name = "MixedUnitary"
    _parnames = ()

    def __init__(self, *args):
        if len(args) == 1:
            rescaledgates = args[0]
            if not isinstance(rescaledgates, list) or not all(
                isinstance(g, RescaledGate) for g in rescaledgates
            ):
                raise ValueError("Expected a list of RescaledGate objects.")
            self.gates = rescaledgates
            self.p = [g.get_scale()**2 for g in self.gates] 
            self.U = [g.get_operation() for g in self.gates]

        elif len(args) == 2:
            probabilities, gates_or_matrices = args
            if len(probabilities) != len(gates_or_matrices):
                raise ValueError(
                    "Lists of probabilities and gates/matrices must have the same length."
                )
            # Check if there are any symbolic probabilities
            contains_symbolic_probabilities = any(
                isinstance(p, (se.Basic, sp.Basic)) for p in probabilities
            )

            if not contains_symbolic_probabilities:
                if not np.allclose(np.sum(probabilities), 1, rtol=1e-13):
                    raise ValueError("Probabilities must sum to 1.")

            self.p = probabilities
            self.gates = []
            self.U = []

            for U, p in zip(gates_or_matrices, probabilities):
                if isinstance(U, (np.ndarray, sp.Matrix, se.Matrix)):
                    # Convert numpy or sympy matrices to symengine and wrap in GateCustom
                    U = _convert_to_symengine(U)
                    if not _is_valid_power_of_2(U.rows):
                        raise ValueError(
                            f"Matrix row count {U.rows} is not a valid power of 2."
                        )
                    gate = GateCustom(U)
                elif isinstance(U, Gate):
                    gate = U
                else:
                    raise ValueError(f"Unsupported type for U: {type(U)}")

                scale = se.sqrt(p)
                self.gates.append(RescaledGate(gate, scale))
                self.U.append(gate)

        else:
            raise ValueError("Invalid number of arguments. Expected 1 or 2.")

        # Use num_qubits() from the Gate class, or calculate from matrix dimensions
        if isinstance(self.gates[0].get_operation(), Gate):
            self.N = self.gates[0].get_operation().num_qubits
        else:
            self.N = int(se.log(self.gates[0].get_operation().rows, 2))

        if self.N < 1:
            raise ValueError("Cannot define a 0-qubit custom mixed unitary channel")
        if self.N > 2:
            raise ValueError(
                "Custom mixed unitary channels larger than 2 qubits are not supported"
            )

        super().__init__()
        self._num_qubits = self.N
     
    def evaluate(self, values: dict):
        """Evaluates symbolic parameters in the MixedUnitary using a dictionary of values.

        Args:
            values (dict): A dictionary where keys are symbolic variables and values are the corresponding numerical values.

        Returns:
            MixedUnitary: A new MixedUnitary instance with evaluated parameters.
        """
        evaluated_p = [p.subs(values) if hasattr(p, 'subs') else p for p in self.p]

        if all(isinstance(prob, (float, int)) or prob.is_number for prob in evaluated_p):
    
            numeric_probs = [float(prob) if isinstance(prob, se.Basic) else prob for prob in evaluated_p]
            if not np.allclose(np.sum(numeric_probs), 1, rtol=1e-13):
                raise ValueError("Evaluated probabilities must sum to 1.")

        evaluated_U = [u.evaluate(values) if hasattr(u, 'evaluate') else u for u in self.U]

        return MixedUnitary(evaluated_p, evaluated_U)

    def krausmatrices(self):
        probabilities = [se.sqrt(p) for p in self.probabilities()]
        unitary_gates = self.unitarygates()

        # Convert gates to matrices before multiplying
        unitary_matrices = [
            gate.matrix() if hasattr(gate, 'matrix') and callable(gate.matrix) else gate.matrix
            for gate in unitary_gates
        ]

        # Make sure calling matrix() or using matrices
        return [
            prob * matrix for prob, matrix in zip(probabilities, unitary_matrices)
        ]

    def probabilities(self):
        return self.p

    def unitarymatrices(self):
        return self.U

    def unitarygates(self):
        return [g.get_operation() for g in self.gates]

    def unwrappedkrausmatrices(self):
        return self.krausmatrices()

    def krausoperators(self):
        # Get the unitary gates
        gates = self.unitarygates()
        scales = [se.sqrt(p) for p in self.probabilities()]
        return [RescaledGate(gate, scale) for gate, scale in zip(gates, scales)]

    @classmethod
    def ismixedunitary(self):
        return True

    def __str__(self):
        """Compact string representation for the MixedUnitary."""
        return self._format_mixedunitary(compact=True)

    def __repr__(self):
        """Full string representation for the MixedUnitary."""
        return self._format_mixedunitary(compact=False)

    def _format_mixedunitary(self, compact):
        """Helper function to format MixedUnitary based on compactness."""
        ps = self.probabilities()
        Us = self.unitarygates()

        repr_string = f"{self._name}("
        sep = "," if compact else ", "

        for i, (prob, gate) in enumerate(zip(ps, Us)):
            # Check if prob is numeric and can be rounded
            if isinstance(prob, (float, int)):
                prob = round(prob, 10)
            # Otherwise, keep the symbolic representation as is

            if isinstance(gate, mc.GateCustom):
                matrix_str = f"Custom([{'; '.join([' '.join(map(str, row)) for row in gate.matrix.tolist()])}])"
                repr_string += f'({prob}, "{matrix_str}")'
            else:
                repr_string += f'({prob}, "{str(gate)}")'

            if i < len(ps) - 1:
                repr_string += sep

        repr_string += ")"
        return repr_string


__all__ = [
    "MixedUnitary",
]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.utils import control_one_defined
import mimiqcircuits as mc
from symengine import pi


class GateSX(mc.Power):
    r"""Single qubit :math:`\sqrt{X}` gate.

    **Matrix representation:**

    .. math::
        \sqrt{\operatorname{X}} = \frac{1}{2} \begin{pmatrix}
            1+i & 1-i \\
            1-i & 1+i\\
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateSX()
        SX
        >>> GateSX().matrix()
        [0.5 + 0.5*I, 0.5 - 0.5*I]
        [0.5 - 0.5*I, 0.5 + 0.5*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateSX(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── SX @ q[0]
        <BLANKLINE>
        >>> GateSX().power(2), GateSX().inverse()
        (X, SX†)
        >>> GateSX().decompose()
        1-qubit circuit with 4 instructions:
        ├── S† @ q[0]
        ├── H @ q[0]
        ├── S† @ q[0]
        └── U(0, 0, 0, (1/4)*pi) @ q[0]
        <BLANKLINE>
    """

    _name = "SX"

    name = "SX"

    def __init__(self):
        super().__init__(mc.GateX(), 1 / 2)

    def inverse(self):
        return GateSXDG()

    def isopalias(self):
        return True

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCSX())

    def _power(self, p):
        # SX * SX = X
        if p % 4 == 2:
            return mc.GateX()

        # (SX * SX) * (SX * SX) = X * X = ID
        if p % 4 == 0:
            return mc.GateID()

        # (SX * SX * SX) * SX = ID => SX * SX * SX = SX†
        if p % 4 == 3:
            return mc.GateSXDG()

        # SX * SX * SX * SX = ID => SX^(n*4) * SX = ID * SX = SX
        if p % 4 == 1:
            return self

        return mc.Power(self, p)

    def __str__(self):
        return f"{self.name}"

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateSDG(), q)
        circ.push(mc.GateH(), q)
        circ.push(mc.GateSDG(), q)
        circ.push(mc.GateU(0, 0, 0, pi / 4), q)
        return circ


class GateSXDG(mc.Inverse):
    r"""Single qubit :math:`\sqrt{X}^\dagger` gate (conjugate transpose of the :math:`\sqrt{X}` gate).

    **Matrix representation:**

    .. math::
        \sqrt{\operatorname{X}}^\dagger = \frac{1}{2} \begin{pmatrix}
            1-i & 1+i \\
            1+i & 1-i\\
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateSXDG()
        SX†
        >>> GateSXDG().matrix()
        [0.5 - 0.5*I, 0.5 + 0.5*I]
        [0.5 + 0.5*I, 0.5 - 0.5*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateSXDG(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── SX† @ q[0]
        <BLANKLINE>
        >>> GateSXDG().power(2), GateSXDG().inverse()
        (SX†**2, SX)
        >>> GateSXDG().decompose()
        1-qubit circuit with 4 instructions:
        ├── S @ q[0]
        ├── H @ q[0]
        ├── S @ q[0]
        └── U(0, 0, 0, (-1/4)*pi) @ q[0]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(GateSX())

    def inverse(self):
        return GateSX()

    def isopalias(self):
        return True

    def _control(self, n):
        return control_one_defined(n, self, mc.GateCSXDG())

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(mc.GateS(), q)
        circ.push(mc.GateH(), q)
        circ.push(mc.GateS(), q)
        circ.push(mc.GateU(0, 0, 0, -pi / 4), q)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.control as mctrl
from mimiqcircuits.operations.gates.standard.pauli import GateX, GateY, GateZ
from mimiqcircuits.operations.gates.standard.s import GateS, GateSDG
from mimiqcircuits.operations.gates.standard.hadamard import GateH
import mimiqcircuits as mc


class GateCX(mctrl.Control):
    r"""Two qubit Controlled-X gate (or CNOT).

    By convention, the first qubit is the control and the second is
    the target

    **Matrix representation:**

    .. math::
        \operatorname{CX} = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 1 \\
            0 & 0 & 1 & 0
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCX(), GateCX().num_controls, GateCX().num_targets
        (CX, 1, 1)
        >>> GateCX().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 0, 1.0]
        [0, 0, 1.0, 0]
        <BLANKLINE>
        >>> c = Circuit().push(GateCX(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CX @ q[0], q[1]
        <BLANKLINE>
        >>> GateCX().power(2), GateCX().inverse()
        (CID, CX)
        >>> GateCX().decompose()
        2-qubit circuit with 1 instructions:
        └── CX @ q[0], q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateX())

    def _decompose(self, circ, qubits, bits, zvars):
        c, t = qubits
        circ.push(mc.GateCX(), c, t)
        return circ


class GateCY(mctrl.Control):
    r"""Two qubit Controlled-Y gate.

    By convention, the first qubit is the control and the second is
    the target

    **Matrix representation:**

    .. math::
        \operatorname{CY} = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 0 & -i \\
            0 & 0 & i & 0
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCY(), GateCY().num_controls, GateCY().num_targets
        (CY, 1, 1)
        >>> GateCY().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 0, -0.0 - 1.0*I]
        [0, 0, 0.0 + 1.0*I, 0]
        <BLANKLINE>
        >>> c = Circuit().push(GateCY(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CY @ q[0], q[1]
        <BLANKLINE>
        >>> GateCY().power(2), GateCY().inverse()
        (CID, CY)
        >>> GateCY().decompose()
        2-qubit circuit with 3 instructions:
        ├── S† @ q[1]
        ├── CX @ q[0], q[1]
        └── S @ q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateY())

    def _decompose(self, circ, qubits, bits, zvars):
        c, t = qubits
        circ.push(GateSDG(), t)
        circ.push(GateCX(), c, t)
        circ.push(GateS(), t)
        return circ


class GateCZ(mctrl.Control):
    r"""Two qubit Controlled-Z gate.

    By convention, the first qubit is the control and the second is
    the target

    **Matrix representation:**

    .. math::
        \operatorname{CZ} = \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            0 & 0 & 0 & -1
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCZ(), GateCZ().num_controls, GateCZ().num_targets
        (CZ, 1, 1)
        >>> GateCZ().matrix()
        [1.0, 0, 0, 0]
        [0, 1.0, 0, 0]
        [0, 0, 1.0, 0]
        [0, 0, 0, -1.0]
        <BLANKLINE>
        >>> c = Circuit().push(GateCZ(), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── CZ @ q[0], q[1]
        <BLANKLINE>
        >>> GateCZ().power(2), GateCZ().inverse()
        (CID, CZ)
        >>> GateCZ().decompose()
        2-qubit circuit with 3 instructions:
        ├── H @ q[1]
        ├── CX @ q[0], q[1]
        └── H @ q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateZ())

    def _decompose(self, circ, qubits, bits, zvars):
        c, t = qubits
        circ.push(GateH(), t)
        circ.push(GateCX(), c, t)
        circ.push(GateH(), t)
        return circ



#
# Copyright © 2022-2023 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits import Operation


class BondDim(Operation):
    r"""Operation to get the bond dimension between two halves of the system and store it in a z-register.

    The bond dimension is only defined for a matrix-product state (MPS), which
    can be written as:

    **State Representation**

    .. math::
        |\psi \rangle = \sum_{s_1,s_2,\ldots=1}^2
        \sum_{i_1}^{\chi_1} \sum_{i_2}^{\chi_2} \ldots \sum_{i_N}^{\chi_N}
        A^{(s_1)}_{i_0i_1} A^{(s_2)}_{i_1 i_2} A^{(s_3)}_{i_2 i_3} \ldots
        A^{(s_N)}_{i_{N-1}i_N} | s_1, s_2, s_3, \ldots, s_N \rangle .

    Here, :math:`\chi_k` is the bond dimension, i.e., the dimension of the index :math:`i_k`.
    The first and last bond dimensions are dummies, :math:`\chi_0=\chi_N=1`.
    A bond dimension of 1 means there is no entanglement between the two halves
    of the system.

    See Also:
        :class:`VonNeumannEntropy`, :class:`SchmidtRank`

    Examples:

        When pushing to a circuit, the qubit index `k` that we give will return
        the bond dimension :math:`i_{k-1}` in the above notation. In other words, we
        associate link `k` with qubit `k+1`. For `k=1`, the bond dimension
        returned will always be 1.

        >>> from mimiqcircuits import *
        >>> k = 5
        >>> c = Circuit()
        >>> c.push(BondDim(), k, 1)
        6-qubit circuit with 1 instructions:
        └── BondDim @ q[5], z[1]
        <BLANKLINE>
    """

    _name = "BondDim"
    _num_zvars = 1
    _num_bits = 0
    _num_zregs = 1
    _num_qubits = 1

    def __init__(self):
        super().__init__()
        self._qregsizes = [self._num_qubits]
        self._zregsizes = [self._num_zvars]
        self._parnames = ()

    def __str__(self):
        return f"{self.name}"

    @property
    def num_qubits(self):
        return self._num_qubits

    @property
    def parnames(self):
        return self._parnames

    def iswrapper(self):
        return False

    def inverse(self):
        return BondDim()

    def power(self, p):
        raise TypeError("BondDim^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled BondDim is not defined.")
    
    @staticmethod
    def isunitary():
        return True


class VonNeumannEntropy(Operation):
    r"""Operation to get the bipartite Von Neumann entanglement entropy and store it in a z-register.

    The entanglement entropy for a bipartition into subsystems :math:`A` and :math:`B`
    is defined for a pure state :math:`\rho = | \psi \rangle\langle \psi |` as:

    **Entanglement Entropy**

    .. math::
        \mathcal{S}(\rho_A) = - \mathrm{Tr}(\rho_A \log_2 \rho_A)
        = - \mathrm{Tr}(\rho_B \log_2 \rho_B)
        = \mathcal{S}(\rho_A)

    where :math:`\rho_A = \mathrm{Tr}_B(\rho)` is the reduced density matrix.
    A product state has :math:`\mathcal{S}(\rho_A)=0` and a maximally entangled state
    between :math:`A` and :math:`B` gives :math:`\mathcal{S}(\rho_A)=1`.

    We only consider bipartitions where :math:`A=\{1,\ldots,k-1\}` and :math:`B=\{k,\ldots,N\}`,
    for some :math:`k` and where :math:`N` is the total number of qubits.

    When the system is open (i.e., with noise) and we are using quantum trajectories,
    the entanglement entropy of each trajectory is returned during execution.

    See Also:
        :class:`BondDim`, :class:`SchmidtRank`

    Examples:

        When pushing to a circuit, the qubit index `k` takes the role of the above bipartition
        into `A` and `B`. For `k=1`, `A` is empty and the entanglement entropy will
        always return 0.

        >>> from mimiqcircuits import *
        >>> k = 5
        >>> c = Circuit()
        >>> c.push(VonNeumannEntropy(), k, 1)
        6-qubit circuit with 1 instructions:
        └── VonNeumannEntropy @ q[5], z[1]
        <BLANKLINE>
    """

    _name = "VonNeumannEntropy"
    _num_zvars = 1
    _num_bits = 0
    _num_zregs = 1
    _num_qubits = 1

    def __init__(self):
        super().__init__()
        self._qregsizes = [self._num_qubits]
        self._zregsizes = [self._num_zvars]
        self._parnames = ()

    def __str__(self):
        return f"{self.name}"

    @property
    def num_qubits(self):
        return self._num_qubits

    @property
    def parnames(self):
        return self._parnames

    def iswrapper(self):
        return False

    def inverse(self):
        return VonNeumannEntropy()

    def power(self, p):
        raise TypeError("VonNeumannEntropy^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled VonNeumannEntropy is not defined.")
    
    @staticmethod
    def isunitary():
        return True


class SchmidtRank(Operation):
    r"""Operation to get the Schmidt rank of a bipartition and store it in a z-register.

    A Schmidt decomposition for a bipartition into subsystems :math:`A` and :math:`B`
    is defined for a pure state as:

    **Schmidt Decomposition**

    .. math::
        |\psi\rangle = \sum_{i=1}^{r} s_i |\alpha_i\rangle \otimes |\beta_i\rangle,

    where :math:`|\alpha_i\rangle` (:math:`|\beta_i\rangle`) are orthonormal states acting
    on :math:`A` (:math:`B`). The Schmidt rank is the number of terms :math:`r` in the sum.
    A product state gives :math:`r=1`, and :math:`r>1` signals entanglement.

    We only consider bipartitions where :math:`A=\{1,\ldots,k-1\}` and :math:`B=\{k,\ldots,N\}`,
    for some :math:`k` and where :math:`N` is the total number of qubits.

    In MPS (Matrix Product States), when the state is optimally compressed, the Schmidt rank should
    be equal to the bond dimension (see :class:`BondDim`).

    See Also:
        :class:`VonNeumannEntropy`, :class:`BondDim`

    Examples:

        When pushing to a circuit, the qubit index `k` takes the role of the above bipartition
        into `A` and `B`. For `k=1`, `A` is empty and the Schmidt rank will
        always return 1.

        >>> from mimiqcircuits import *
        >>> k = 5
        >>> c = Circuit()
        >>> c.push(SchmidtRank(), k, 1)
        6-qubit circuit with 1 instructions:
        └── SchmidtRank @ q[5], z[1]
        <BLANKLINE>
    """

    _name = "SchmidtRank"
    _num_zvars = 1
    _num_bits = 0
    _num_zregs = 1
    _num_qubits = 1

    def __init__(self):
        super().__init__()
        self._qregsizes = [self._num_qubits]
        self._zregsizes = [self._num_zvars]
        self._parnames = ()

    def __str__(self):
        return f"{self.name}"

    @property
    def num_qubits(self):
        return self._num_qubits

    @property
    def parnames(self):
        return self._parnames

    def iswrapper(self):
        return False

    def inverse(self):
        return SchmidtRank()

    def power(self, p):
        raise TypeError("SchmidtRank^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled SchmidtRank is not defined.")
    
    @staticmethod
    def isunitary():
        return True


__all__ = ["BondDim", "VonNeumannEntropy", "SchmidtRank"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.operation import Operation
import mimiqcircuits.lazy as lz


class Barrier(Operation):
    """Barrier operation.

    A barrier is a special operation that does not affect the quantum state or
    the execution of a circuit, but it prevents compression or optimization
    operation from being applied across it.

    Examples:

        Adding Barrier operation to the Circuit (The args can be: range, list,
        tuple, set or int)

        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Barrier(1), 1)
        2-qubit circuit with 1 instructions:
        └── Barrier @ q[1]
        <BLANKLINE>

        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Barrier(1), range(0,4))
        4-qubit circuit with 4 instructions:
        ├── Barrier @ q[0]
        ├── Barrier @ q[1]
        ├── Barrier @ q[2]
        └── Barrier @ q[3]
        <BLANKLINE>

        Adding Barrier to the circuit as a multi-qubits gate

        >>> from mimiqcircuits import *
        >>> c= Circuit()
        >>> c.push(Barrier(5),1,2,3,4,5)
        6-qubit circuit with 1 instructions:
        └── Barrier @ q[1,2,3,4,5]
        <BLANKLINE>
    """

    _name = "Barrier"
    _num_qubits = None
    _num_bits = 0
    _num_zvars = 0

    def __init__(self, num_qubits):
        """
        Initialize a barrier operation

        Args:
            num_qubits: number of qubits the barrier will cover

        Raises:
            ValueError: if num_qubits is less than 1
        """
        super().__init__()

        if num_qubits < 1:
            raise ValueError("Number of qubits must be greater than 0")

        self._num_qubits = num_qubits
        self._qregsizes = [num_qubits]

    def __new__(self, *args):
        if len(args) == 0:
            return lz.LazyExpr(Barrier, lz.LazyArg())
        elif len(args) == 1:
            return object.__new__(self)
        else:
            raise ValueError("Invalid number of arguments.")

    def inverse(self):
        return self

    def power(self, p):
        raise TypeError("Barrier^p is not defined.")

    def control(self, num_qubits):
        raise TypeError("Controlled Barrier is not defined.")

    def iswrapper(self):
        return False

    def asciiwidth(self, qubits, bits, zvars):
        return 1
    
    @staticmethod
    def isunitary():
        return True


# export operations
__all__ = ["Barrier"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from bitarray import bitarray, frozenbitarray


def _helper_bitvec_to_int(arr):
    res = 0
    for bit in arr:
        res = (res << 1) | bit
    return res


# Function to convert a bitarray to an integer
def bitvec_to_int(arr):
    if not isinstance(arr, bitarray) or not isinstance(arr, frozenbitarray):
        raise TypeError("Input must be a bitarray or BitVector.")
    num_bits = len(arr)
    if num_bits > 0:
        return _helper_bitvec_to_int(arr)
    else:
        raise ValueError("Input array is empty.")


# Function to convert an integer to a bitarray of given size (pad)
def int_to_bitvec(x, pad, frozen=True):
    if not isinstance(x, int):
        raise TypeError("Input must be an integer.")
    if x < 0:
        raise ValueError("Input integer must be non-negative.")

    binary_str = bin(x)[2:].zfill(pad)

    if frozen:
        return frozenbitarray(binary_str)
    else:
        return bitarray(binary_str)


class BitString:
    """BitString for the quantum states.

    Representation of the quantum state of a quantum register with definite
    values for each qubit.

    Examples:

        Initialization:

        >>> from mimiqcircuits import *
        >>> from bitarray import bitarray
        >>> BitString(16) # number of qubits
        bs"0000000000000000"
        >>> BitString('10101') # binary string
        bs"10101"
        >>> BitString([1,0,0,0,1]) # binary string
        bs"10001"
        >>> BitString((1,0,0,0,1)) # binary string
        bs"10001"

        >>> BitString(bitarray('101010')) # bitarray
        bs"101010"

        Other initializations:

        >>> BitString.fromnonzeros(16, [1, 3, 5, 7, 9, 11, 13, 15])
        bs"0101010101010101"
        >>> BitString.fromfunction(16, lambda i: i % 2 == 1)
        bs"0101010101010101"
        >>> BitString.fromstring('10101')
        bs"10101"
        >>> BitString.fromint(16, 21)
        bs"1010100000000000"
        >>> BitString.fromint(16, 21, 'little')
        bs"0000000000010101"

        Accessing the bits:

        >>> bs = BitString(16)
        >>> bs[0] # get the 0th bit
        0
        >>> bs[0:4] # get the first 4 bits
        bs"0000"

        Bitwise operations:

        >>> bs1 = BitString('10101')
        >>> bs2 = BitString('11100')
        >>> bs1 | bs2 # OR
        bs"11101"
        >>> bs1 & bs2 # AND
        bs"10100"
        >>> bs1 ^ bs2 # XOR
        bs"01001"
        >>> ~bs1 # NOT
        bs"01010"
        >>> bs1 << 2 # left shift
        bs"10100"
        >>> bs1 >> 2 # right shift
        bs"00101"

        Other operations:

        >>> bs1 + bs2 # concatenation
        bs"1010111100"
        >>> bs1 * 2 # repetition
        bs"1010110101"
    """

    _bits = bitarray(0)

    def __init__(self, arg, indices=None):
        """Initialize the BitString.

        If the number of qubits is given, then the BitString is initialized to
        the all-zero state.
        If a binary string or a bitarray is given, then the BitString is the
        corresponding state.

        Examples:
            >>> BitString(16) # number of qubits
            bs"0000000000000000"
            >>> BitString('10101') # binary string
            bs"10101"
            >>> BitString(bitarray('101010')) # bitarray
            bs"101010"
        """
        if isinstance(arg, int):
            if indices is None:
                bitstring = arg * "0"
            else:
                if not all(0 <= i < arg for i in indices):
                    raise ValueError("Index out of range for the given number of qubits.")
                bitstring = "".join("1" if i in indices else "0" for i in range(arg))
        elif isinstance(arg, (str, bitarray, frozenbitarray, list, tuple)):
            bitstring = arg
        else:
            raise TypeError(
                "Invalid input type. Expected 'str', 'int', 'bitarray', 'list', 'tuple', 'frozenbitarray'"
            )

        self._bits = frozenbitarray(bitstring)

    @property
    def bits(self):
        return self._bits

    @bits.setter
    def bits(self, bits):
        raise AttributeError("Cannot set the bits of a BitString.")

    @staticmethod
    def fromnonzeros(num_qubits: int, nonzeros: list):
        """Initialize a BitString with specific non-zero qubits.

        Arguments:
            num_qubits (int): The number of qubits in the BitString.
            nonzeros (list): A list of non-zero qubit indices to set in the BitString.

        Returns:
            A BitString with the specified non-zero qubits.
        """
        bitstring = ""

        if not all(0 <= i < num_qubits for i in nonzeros):
            raise ValueError("Invalid nonzero index.")

        for i in range(num_qubits):
            bitstring += "1" if i in nonzeros else "0"

        return BitString(bitstring)

    @staticmethod
    def fromfunction(num_qubits: int, f: type[lambda: None]):
        """Initialize a BitString from a function.

        Arguments:
            num_qubits (int): The number of qubits in the BitString.
            f (function): A function that takes an integer and returns a boolean.

        Returns:
            A BitString.
        """
        bitstring = ""
        for i in range(num_qubits):
            bitstring += "1" if f(i) else "0"
        return BitString(bitstring)

    @staticmethod
    def fromstring(bitstring: str):
        """Initialize a BitString from a string.

        Arguments:
            bitstring (str): The string representation of the BitString.

        Returns:
            A BitString.
        """
        return BitString(bitstring)

    @staticmethod
    def fromint(num_qubits: int, integer: int, endianess: str = "big"):
        """Initialize a BitString from an integer.

        Arguments:
            num_qubits (int): The number of qubits in the BitString.
            integer (int): The integer value of the BitString.
            endianess (str): The endianess of the integer. Default is 'big'.

        Returns:
            A BitString.
        """
        if len(bin(integer)[2:]) > num_qubits:
            raise ValueError("Integer is too large for the given number of qubits.")

        if endianess == "little":
            bitstring = bin(integer)[2:].zfill(num_qubits)[::-1]
        elif endianess == "big":
            bitstring = bin(integer)[2:].zfill(num_qubits)
        else:
            raise ValueError("endian must be either 'big' or 'little'")

        return BitString(bitstring[::-1])

    def num_qubits(self):
        """Return the number of qubits in the BitString."""
        return len(self.bits)

    def nonzeros(self):
        """Return the indices of the non-zero qubits."""
        return [i for i, bit in enumerate(self.bits) if bit]

    def zeros(self):
        """Return the indices of the zero qubits."""
        return [i for i, bit in enumerate(self.bits) if not bit]

    def tointeger(self, endianess: str = "big"):
        """Return the integer value of the BitString.

        Arguments:
            endianess (str): The endianess of the integer. Default is 'big'.

        Returns:
            The integer value of the BitString.
        """
        if endianess == "big":
            return bitvec_to_int(self.bits[::-1])
        elif endianess == "little":
            return bitvec_to_int(self.bits)
        else:
            raise ValueError("Invalid endianess. Must be 'big' or 'little'.")

    def to01(self, endianess="big"):
        """Return the binary string representation of the BitString.

        Arguments:
            endianess (str): The endianess of the integer. Default is 'big'

        Retruns:
            The binary string representation of the BitString.
        """
        if endianess == "big":
            return "".join(map(str, self.bits))
        elif endianess == "little":
            return "".join(map(str, reversed(self.bits)))
        else:
            raise ValueError("Invalid endianess. Must be 'big' or 'little'.")

    def toindex(self, endianess: str = "big"):
        """Return the integer index of the BitString.

        Arguments:
            endianess (str): The endianess of the integer. Default is 'big'.

        Returns:
            The integer index of the BitString.
        """
        if endianess == "big":
            return bitvec_to_int(self.bits[::-1])
        elif endianess == "little":
            return bitvec_to_int(self.bits)
        else:
            raise ValueError("endian must be either 'big' or 'little'")

    def __eq__(self, other):
        if not isinstance(other, BitString):
            return False
        return self.bits == other.bits

    def __str__(self):
        s = f"{len(self.bits)}-qubit BitString"
        nz = self.nonzeros()
        if len(nz) != 0:
            s += f" with {len(nz)} non-zero qubits:\n"
            s += f"├── |{self.bits.to01()}⟩\n"
            s += f"└── non-zero qubits: {nz}"
        else:
            s += f":\n└── |{self.bits.to01()}⟩"
        return s

    def __repr__(self):
        return "bs" + '"' + self.to01() + '"'

    def __len__(self):
        return len(self.bits)

    def __hash__(self):
        return hash(self.bits.to01())

    def __iter__(self):
        return iter(self.bits)

    def __getitem__(self, index):
        if type(index) is slice:
            return BitString(self.bits[index])
        return self.bits[index]
    
    def __setitem__(self, index, value):
        """Set a specific bit by creating a new frozenbitarray."""
        if not isinstance(value, (bool, int)) or value not in (0, 1):
            raise ValueError("Value must be a boolean or an integer (0 or 1).")
        
        # Convert frozenbitarray to a mutable bitarray
        temp_bits = self._bits.tolist()
        temp_bits[index] = value
        
        # Recreate the frozenbitarray with the updated bits
        self._bits = frozenbitarray(temp_bits)

    def __or__(self, other):
        if not isinstance(other, BitString):
            raise TypeError("BitString can only be ORed with another BitString.")
        return BitString(self.bits | other.bits)

    def __and__(self, other):
        if not isinstance(other, BitString):
            raise TypeError("BitString can only be ANDed with another BitString.")
        return BitString(self.bits & other.bits)

    def __xor__(self, other):
        if not isinstance(other, BitString):
            raise TypeError("BitString can only be XORed with another BitString.")
        return BitString(self.bits ^ other.bits)

    def __invert__(self):
        return BitString(~self.bits)

    def __lshift__(self, other):
        if not isinstance(other, int):
            raise TypeError("BitString can only be left shifted by an integer.")
        return BitString(self.bits << other)

    def __rshift__(self, other):
        if not isinstance(other, int):
            raise TypeError("BitString can only be right shifted by an integer.")
        return BitString(self.bits >> other)

    def __add__(self, other):
        if not isinstance(other, BitString):
            raise TypeError("BitString can only be added with another BitString.")
        return BitString(self.bits + other.bits)

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError("BitString can only be multiplied by an integer.")
        return BitString(self.bits * other)


__all__ = ["BitString"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.control as mctrl
from mimiqcircuits.operations.gates.standard.swap import GateSWAP
from mimiqcircuits.operations.gates.standard.cpauli import GateCX
from mimiqcircuits.operations.gates.standard.cnx import GateCCX


class GateCSWAP(mctrl.Control):
    r"""Three qubit Controlled-SWAP gate.

    By convention, the first qubit is the control and last two are the
    targets.

    Examples:
        >>> from mimiqcircuits import *
        >>> GateCSWAP(), GateCSWAP().num_controls, GateCSWAP().num_targets, GateCSWAP().num_qubits
        (CSWAP, 1, 2, 3)
        >>> GateCSWAP().matrix()
        [1.0, 0, 0, 0, 0, 0, 0, 0]
        [0, 1.0, 0, 0, 0, 0, 0, 0]
        [0, 0, 1.0, 0, 0, 0, 0, 0]
        [0, 0, 0, 1.0, 0, 0, 0, 0]
        [0, 0, 0, 0, 1.0, 0, 0, 0]
        [0, 0, 0, 0, 0, 0, 1.0, 0]
        [0, 0, 0, 0, 0, 1.0, 0, 0]
        [0, 0, 0, 0, 0, 0, 0, 1.0]
        <BLANKLINE>
        >>> c = Circuit().push(GateCSWAP(), 0, 1, 2)
        >>> GateCSWAP().power(2), GateCSWAP().inverse()
        (C(⨷ ² ID), CSWAP)
        >>> c = Circuit().push(GateCSWAP(), 0, 1, 2)
        >>> c
        3-qubit circuit with 1 instructions:
        └── CSWAP @ q[0], q[1,2]
        <BLANKLINE>
        >>> GateCSWAP().power(2), GateCSWAP().inverse()
        (C(⨷ ² ID), CSWAP)
        >>> GateCSWAP().decompose()
        3-qubit circuit with 3 instructions:
        ├── CX @ q[2], q[1]
        ├── C₂X @ q[0,1], q[2]
        └── CX @ q[2], q[1]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(1, GateSWAP())

    def _decompose(self, circ, qubits, bits, zvars):
        c, t1, t2 = qubits
        circ.push(GateCX(), t2, t1)
        circ.push(GateCCX(), c, t1, t2)
        circ.push(GateCX(), t2, t1)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits.operations.gates.gate as mcg
from mimiqcircuits.matrices import umatrix
from mimiqcircuits.operations.utils import control_one_defined
import mimiqcircuits as mc
from sympy import I, pi, sin, cos, acos, Abs, simplify, exp, Expr, log, Matrix
import numpy as np
import sympy as sp
import symengine as se
from scipy.linalg import expm, logm


class GateU(mcg.Gate):
    r"""Single qubit generic unitary phase gate.


    **Matrix representation:**

    .. math::
        \operatorname{U}(\theta, \phi, \lambda, \gamma) =
        \mathrm{e}^{i\gamma}
        \begin{pmatrix}
            \cos\left(\frac{\theta}{2}\right) & -\mathrm{e}^{i\lambda}\sin\left(\frac{\theta}{2}\right)\\
            \mathrm{e}^{i\phi}\sin\left(\frac{\theta}{2}\right) & \mathrm{e}^{i(\phi+\lambda)}\cos\left (\frac{\theta}{2}\right)
        \end{pmatrix}

    Parameters:
        theta (float): Euler angle 1 in radians.
        phi (float): Euler angle 2 in radians.
        lambda (float): Euler angle 3 in radians.
        gamma (float, optional): Euler angle 4 in radians (default is 0).

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta, phi, lmbda, gamma = symbols('theta phi lambda gamma')
        >>> GateU(theta, phi, lmbda, gamma)
        U(theta, phi, lambda, gamma)
        >>> GateU(theta, phi, lmbda, gamma).matrix()
        [exp(I*gamma)*cos((1/2)*theta), -exp(I*(gamma + lambda))*sin((1/2)*theta)]
        [exp(I*(gamma + phi))*sin((1/2)*theta), exp(I*(gamma + lambda + phi))*cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateU(theta, phi, lmbda, gamma), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── U(theta, phi, lambda, gamma) @ q[0]
        <BLANKLINE>
        >>> GateU(theta, phi, lmbda, gamma).power(2), GateU(theta, phi, lmbda, gamma).inverse()
        (U(theta, phi, lambda, gamma)**2, U(-theta, -lambda, -phi, -gamma))
        >>> GateU(theta, phi, lmbda, gamma).decompose()
        1-qubit circuit with 1 instructions:
        └── U(theta, phi, lambda, gamma) @ q[0]
        <BLANKLINE>
        >>> c = Circuit().push(GateU(theta, phi, lmbda, gamma), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── U(theta, phi, lambda, gamma) @ q[0]
        <BLANKLINE>
        >>> GateU(theta, phi, lmbda, gamma).power(2), GateU(theta, phi, lmbda, gamma).inverse()
        (U(theta, phi, lambda, gamma)**2, U(-theta, -lambda, -phi, -gamma))
        >>> GateU(theta, phi, lmbda, gamma).decompose()
        1-qubit circuit with 1 instructions:
        └── U(theta, phi, lambda, gamma) @ q[0]
        <BLANKLINE>
    """

    _name = "U"

    _num_qubits = 1
    _qregsizes = [1]

    _parnames = ("theta", "phi", "lmbda", "gamma")

    def __init__(self, theta, phi, lmbda, gamma=0.0):
        self.theta = theta
        self.phi = phi
        self.lmbda = lmbda
        self.gamma = gamma

    def _matrix(self):
        return umatrix(self.theta, self.phi, self.lmbda, self.gamma)

    def inverse(self):
        return GateU(-self.theta, -self.lmbda, -self.phi, -self.gamma)

    def _control(self, n):
        return control_one_defined(
            n, self, mc.GateCU(self.theta, self.phi, self.lmbda, self.gamma)
        )

    def _power(self, p):
        
        if self.is_symbolic():
            return mc.Power(self, p)

        def to_numeric(value):
            if isinstance(value, (se.Basic,sp.Basic)) and value == se.pi or value == se.pi:
                return float(value)
            return value

        theta_value = to_numeric(self.theta)
        phi_value = to_numeric(self.phi)
        lambda_value = to_numeric(self.lmbda)
        gamma_value = to_numeric(self.gamma)

        numeric_gate = GateU(theta_value, phi_value, lambda_value, gamma_value)

        matrix = numeric_gate.matrix().tolist()

        matrix_np = np.array(self.convert_to_numeric(matrix))

        pow_matrix = expm(p * logm(matrix_np))

        # Compute the angles based on the resulting matrix and prevent error raising becouse of division by zero.
        with np.errstate(divide='ignore'):
            theta_p = 2 * np.arccos(np.abs(pow_matrix[0, 0]))
            gamma_p = np.angle(pow_matrix[0, 0])
            phi_p = np.angle(pow_matrix[1, 0] / np.sin(theta_p / 2)) - gamma_p
            lambda_p = np.angle(-pow_matrix[0, 1] / np.sin(theta_p / 2)) - gamma_p

        return GateU(theta_p, phi_p, lambda_p, gamma_p)

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(self, q)
        return circ

    def convert_to_numeric(self, matrix):
        """
        Convert a symbolic matrix to a numeric numpy array.
        """
        return np.array([[complex(elem.evalf()) for elem in row] for row in matrix], dtype=np.complex128)



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.__version__ import __version__

from mimiqcircuits.circuit import Circuit

from mimiqcircuits.bitstrings import BitString

from mimiqcircuits.operations.operation import Operation

from mimiqcircuits.operations.control import Control
from mimiqcircuits.operations.parallel import Parallel
from mimiqcircuits.operations.inverse import Inverse
from mimiqcircuits.operations.power import Power
from mimiqcircuits.operations.barrier import Barrier
from mimiqcircuits.operations.ifstatement import IfStatement
from mimiqcircuits.operations.measure import (
    AbstractMeasurement,
    Measure,
    MeasureX,
    MeasureY,
    MeasureZ,
)
from mimiqcircuits.operations.reset import Reset, ResetX, ResetY, ResetZ
from mimiqcircuits.operations.gates.gate import Gate

from mimiqcircuits.operations.gates.custom import GateCustom

from mimiqcircuits.operations.gates.standard.u import GateU
from mimiqcircuits.operations.gates.standard.id import GateID
from mimiqcircuits.operations.gates.standard.pauli import GateX, GateY, GateZ
from mimiqcircuits.operations.gates.standard.hadamard import (
    GateH,
    GateHXY,
    GateHXZ,
    GateHYZ,
)
from mimiqcircuits.operations.gates.standard.s import GateS, GateSDG
from mimiqcircuits.operations.gates.standard.t import GateT, GateTDG
from mimiqcircuits.operations.gates.standard.sx import GateSX, GateSXDG
from mimiqcircuits.operations.gates.standard.sy import GateSY, GateSYDG
from mimiqcircuits.operations.gates.standard.rotations import (
    GateRX,
    GateRY,
    GateRZ,
    GateR,
)
from mimiqcircuits.operations.gates.standard.deprecated import GateU1, GateU2, GateU3

from mimiqcircuits.operations.gates.standard.cpauli import GateCX, GateCY, GateCZ
from mimiqcircuits.operations.gates.standard.chadamard import GateCH
from mimiqcircuits.operations.gates.standard.phase import GateP
from mimiqcircuits.operations.gates.standard.swap import GateSWAP
from mimiqcircuits.operations.gates.standard.iswap import GateISWAP
from mimiqcircuits.operations.gates.standard.cs import GateCS, GateCSDG
from mimiqcircuits.operations.gates.standard.csx import GateCSX, GateCSXDG
from mimiqcircuits.operations.gates.standard.ecr import GateECR
from mimiqcircuits.operations.gates.standard.dcx import GateDCX
from mimiqcircuits.operations.gates.standard.cphase import GateCP
from mimiqcircuits.operations.gates.standard.cu import GateCU
from mimiqcircuits.operations.gates.standard.crotations import GateCRX, GateCRY, GateCRZ
from mimiqcircuits.operations.gates.standard.interactions import (
    GateRXX,
    GateRYY,
    GateRZZ,
    GateRZX,
    GateXXplusYY,
    GateXXminusYY,
)
from mimiqcircuits.operations.gates.standard.cnx import GateCCX, GateC3X
from mimiqcircuits.operations.gates.standard.cnp import GateCCP
from mimiqcircuits.operations.gates.standard.cswap import GateCSWAP

from mimiqcircuits.operations.gates.generalized.qft import QFT
from mimiqcircuits.operations.gates.generalized.phasegradient import PhaseGradient

from mimiqcircuits.instruction import Instruction

from mimiqcircuits.remote import MimiqConnection

from mimiqcircuits.gatedecl import GateCall, GateDecl, gatedecl

from mimiqcircuits.lazy import control, parallel, inverse, power, LazyArg, LazyExpr

from mimiqcircuits.operations.gates.generalized.diffusion import Diffusion

from mimiqcircuits.operations.gates.generalized.polynomialoracle import PolynomialOracle

from mimiqcircuits.qcsresults import QCSResults
from mimiqcircuits.canvas import AsciiCanvas, AsciiCircuit
from mimiqcircuits.operations.operator import AbstractOperator
from mimiqcircuits.operations.gates.generalized.paulistring import PauliString
from mimiqcircuits.operations.noisechannel.standards.ampdamping import (
    AmplitudeDamping,
    GeneralizedAmplitudeDamping,
)

from mimiqcircuits.operations.operators.sigmas import (
    SigmaMinus,
    SigmaPlus,
)
from mimiqcircuits.operations.operators.diagonals import DiagonalOp
from mimiqcircuits.operations.operators.projectors import (
    Projector0,
    Projector00,
    Projector01,
    Projector1,
    Projector10,
    Projector11,
    ProjectorX0,
    ProjectorX1,
    ProjectorY0,
    ProjectorY1,
    ProjectorZ0,
    ProjectorZ1,
)

from mimiqcircuits.operations.noisechannel.standards.depolarizing import (
    Depolarizing,
    Depolarizing1,
    Depolarizing2,
)
from mimiqcircuits.operations.operators.custom import Operator
from mimiqcircuits.operations.noisechannel.standards.phaseamplitudedamping import (
    PhaseAmplitudeDamping,
    ThermalNoise,
)
from mimiqcircuits.operations.noisechannel.kraus import Kraus
from mimiqcircuits.operations.noisechannel.mixedunitary import MixedUnitary
from mimiqcircuits.operations.krauschannel import krauschannel
from mimiqcircuits.operations.noisechannel.standards.pauli import (
    PauliNoise,
    PauliX,
    PauliY,
    PauliZ,
)
from mimiqlink.connection import QPERFECT_CLOUD, QPERFECT_CLOUD2
from mimiqcircuits.operations.amplitude import Amplitude
from mimiqcircuits.operations.expectationvalue import ExpectationValue
from mimiqcircuits.operations.entanglement import (
    SchmidtRank,
    VonNeumannEntropy,
    BondDim,
)
from mimiqcircuits.operations.measurereset import (
    MeasureReset,
    MeasureResetX,
    MeasureResetY,
    MeasureResetZ,
)
from mimiqcircuits.operations.gates.delay import Delay
from mimiqcircuits.operations.noisechannel.standards.projectivenoise import (
    ProjectiveNoise,
    ProjectiveNoiseX,
    ProjectiveNoiseY,
    ProjectiveNoiseZ,
)
from mimiqcircuits.operations.pairmeasure import MeasureXX, MeasureYY, MeasureZZ
from mimiqcircuits.operations.rescaledgates import RescaledGate
from mimiqcircuits.operations.annotations import (
    AbstractAnnotation,
    Detector,
    QubitCoordinates,
    ShiftCoordinates,
    ObservableInclude,
    Tick,
)
from mimiqcircuits.classical.classical_not import Not


class GATES:
    """
    Helper class to list gates.

    **NOTE**: Cannot instantiated called directly. Use `GATES.list()` for a list
    of gate types, and `help(GATES)` for the .


    **Single qubit gates**
        :func:`GateX` :func:`GateY` :func:`GateZ` :func:`GateH`
        :func:`GateS` :func:`GateSDG`
        :func:`GateT` :func:`GateTDG`
        :func:`GateSX` :func:`GateSXDG`
        :func:`GateID`

    **Single qubit gates (parametric)**
        :func:`GateU` :func:`GateP`
        :func:`GateRX` :func:`GateRY` :func:`GateRZ` :func:`GateP`

    **Two qubit gates**
        :func:`GateCX` :func:`GateCY` :func:`GateCZ`
        :func:`GateCH`
        :func:`GateSWAP` :func:`GateISWAP`
        :func:`GateCS` :func:`GateCSX`
        :func:`GateECR` :func:`GateDCX`

    **Two qubit gates (parametric)**
        :func:`GateCU`
        :func:`GateCP`
        :func:`GateCRX` :func:`GateCRY` :func:`GateCRZ`
        :func:`GateRXX` :func:`GateRYY` :func:`GateRZZ`
        :func:`GateXXplusYY` :func:`GateXXminusYY`

    **Other**
        :func:`GateCustom`
    """

    def __new__(cls):
        raise NotImplementedError

    @staticmethod
    def list():
        return [
            GateCustom,
            GateU,
            GateID,
            GateX,
            GateY,
            GateZ,
            GateH,
            GateHXY,
            GateHXZ,
            GateHYZ,
            GateS,
            GateSDG,
            GateT,
            GateTDG,
            GateSX,
            GateSXDG,
            GateSY,
            GateSYDG,
            GateRX,
            GateRY,
            GateRZ,
            GateR,
            GateU1,
            GateU2,
            GateU3,
            GateCX,
            GateCY,
            GateCZ,
            GateCH,
            GateSWAP,
            GateISWAP,
            GateCS,
            GateCSDG,
            GateCSX,
            GateCSXDG,
            GateECR,
            GateDCX,
            GateCP,
            GateCU,
            GateCRX,
            GateCRY,
            GateCRZ,
            GateRXX,
            GateRYY,
            GateRZZ,
            GateRZX,
            GateXXplusYY,
            GateXXminusYY,
            GateCCX,
            GateC3X,
            GateCCP,
            GateCSWAP,
            GateP,
        ]


# Export specific classes, and functions.
__all__ = [
    "Circuit",
    "BitString",
    "Operation",
    "Control",
    "Parallel",
    "Inverse",
    "Power",
    "Barrier",
    "IfStatement",
    "AbstractMeasurement",
    "Measure",
    "MeasureX",
    "MeasureY",
    "MeasureZ",
    "Reset",
    "ResetX",
    "ResetY",
    "ResetZ",
    "MeasureReset",
    "MeasureResetX",
    "MeasureResetY",
    "MeasureResetZ",
    "Gate",
    "GateCustom",
    "GateU",
    "GateID",
    "GateX",
    "GateY",
    "GateZ",
    "GateH",
    "GateHXY",
    "GateHXZ",
    "GateHYZ",
    "GateS",
    "GateSDG",
    "GateT",
    "GateTDG",
    "GateSX",
    "GateSXDG",
    "GateSY",
    "GateSYDG",
    "GateRX",
    "GateRY",
    "GateRZ",
    "GateR",
    "GateU1",
    "GateU2",
    "GateU3",
    "GateCX",
    "GateCY",
    "GateCZ",
    "GateCH",
    "GateSWAP",
    "GateISWAP",
    "GateCS",
    "GateCSDG",
    "GateCSX",
    "GateCSXDG",
    "GateECR",
    "GateDCX",
    "GateCP",
    "GateCU",
    "GateCRX",
    "GateCRY",
    "GateCRZ",
    "GateRXX",
    "GateRYY",
    "GateRZZ",
    "GateRZX",
    "GateXXplusYY",
    "GateXXminusYY",
    "GateCCX",
    "GateC3X",
    "GateCCP",
    "GateCSWAP",
    "QFT",
    "PhaseGradient",
    "Instruction",
    "GateP",
    "MimiqConnection",
    "GateCall",
    "GateDecl",
    "gatedecl",
    "control",
    "parallel",
    "inverse",
    "power",
    "LazyArg",
    "LazyExpr",
    "Diffusion",
    "PolynomialOracle",
    "QCSResults",
    "AsciiCanvas",
    "AsciiCircuit",
    "AbstractOperator",
    "PauliString",
    "AmplitudeDamping",
    "GeneralizedAmplitudeDamping",
    "SigmaMinus",
    "SigmaPlus",
    "DiagonalOp",
    "Projector0",
    "Projector00",
    "Projector01",
    "Projector1",
    "Projector10",
    "Projector11",
    "Depolarizing",
    "Depolarizing1",
    "Depolarizing2",
    "ProjectorX0",
    "ProjectorX1",
    "ProjectorY0",
    "ProjectorY1",
    "ProjectorZ0",
    "ProjectorZ1",
    "Operator",
    "PauliNoise",
    "krauschannel",
    "PauliX",
    "PauliY",
    "PauliZ",
    "ThermalNoise",
    "PhaseAmplitudeDamping",
    "Kraus",
    "MixedUnitary",
    "QPERFECT_CLOUD",
    "QPERFECT_CLOUD2",
    "Amplitude",
    "ExpectationValue",
    "BondDim",
    "VonNeumannEntropy",
    "SchmidtRank",
    "Delay",
    "ProjectiveNoise",
    "ProjectiveNoiseX",
    "ProjectiveNoiseY",
    "ProjectiveNoiseZ",
    "MeasureZZ",
    "MeasureXX",
    "MeasureYY",
    "RescaledGate",
    "AbstractAnnotation",
    "Detector",
    "QubitCoordinates",
    "ShiftCoordinates",
    "ObservableInclude",
    "Tick",
    "Not",
    "GATES",
]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from mimiqcircuits.proto.qcsrproto import toproto_qcsr, fromproto_qcsr

from mimiqcircuits.bitstrings import bitvec_to_int
from statistics import mean, median, stdev


class QCSResults:
    """
    Represents the results of quantum computations obtained from quantum cloud services (QCS).

    Args:
        simulator (str): The name of the quantum simulator.
        version (str): The version of the quantum simulator.
        fidelities (list): List of fidelity estimates from different executions.
        avggateerrors (list): List of average multi-qubit gate errors from different executions.
        cstates (list): List of classical states obtained from executions.
        zstates (list): Not used in the current implementation.
        amplitudes (dict): Dictionary of statevector amplitudes for different quantum states.
        timings (dict): Dictionary of timing information for different phases of the computation.

    """

    def __init__(
        self,
        simulator=None,
        version=None,
        fidelities=None,
        avggateerrors=None,
        cstates=None,
        zstates=None,
        amplitudes=None,
        timings=None,
    ):
        self.simulator = simulator
        self.version = version
        self.fidelities = fidelities if fidelities is not None else []
        self.avggateerrors = avggateerrors if avggateerrors is not None else []
        self.cstates = cstates if cstates is not None else []
        self.zstates = zstates if zstates is not None else []
        self.amplitudes = amplitudes if amplitudes is not None else {}
        self.timings = timings if timings is not None else {}

    def __repr__(self):
        result_str = "QCSResults:\n"

        if self.simulator is not None:
            result_str += f"├── simulator: {self.simulator} {self.version}\n"

        # filter out the ones < 1e-7
        timings = {k: v for k, v in self.timings.items() if v > 1e-7}

        result_str += "├── timings:\n"
        for key, value in list(timings.items())[:-1]:
            result_str += f"│    ├── {key} time: {value}s\n"

        key, value = list(timings.items())[-1]
        result_str += f"│    └── {key} time: {value}s\n"

        if len(self.fidelities) == 1:
            result_str += f"├── fidelity estimate: {self.fidelities[0]:.3g}\n"
            result_str += f"├── average multi-qubit gate error estimate: {self.avggateerrors[0]:.3g}\n"
        elif len(self.fidelities) != 0:
            result_str += "├── fidelity estimate:\n"
            result_str += f"│    ├── min, max: {min(self.fidelities):.3g}, {max(self.fidelities):.3g}\n"
            result_str += f"│    ├── mean: {mean(self.fidelities):.3g}\n"
            result_str += f"│    ├── median: {median(self.fidelities):.3g}\n"
            result_str += f"│    └── std: {stdev(self.fidelities):.3g}\n"
            result_str += "├── average multi-qubit gate error estimate:\n"
            result_str += f"│    ├── min, max: {min(self.avggateerrors):.3g}, {max(self.avggateerrors):.3g}\n"
            result_str += f"│    ├── mean: {mean(self.avggateerrors):.3g}\n"
            result_str += f"│    ├── median: {median(self.avggateerrors):.3g}\n"
            result_str += f"│    └── std: {stdev(self.avggateerrors):.3g}\n"

        if len(self.cstates) != 0:
            hist = self.histogram()
            outcomes = sorted(hist, key=hist.get, reverse=True)[0:5]
            result_str += "├── most sampled:\n"
            for bs in outcomes[:-1]:
                result_str += f'│    ├── bs"{bs.to01()}" => {hist[bs]}\n'

            bs = outcomes[-1]
            result_str += f'│    └── bs"{bs.to01()}" => {hist[bs]}\n'

        if len(self.zstates) > 0 and any(len(zstate) > 0 for zstate in self.zstates):
            result_str += "├── zreg (most sampled):\n"
            z_hist = self.histzvars()
            for zstate, count in list(z_hist.items())[:-1]:
                result_str += f"│    ├── [{', '.join(map(str, zstate))}] => {count}\n"
            zstate, count = list(z_hist.items())[-1]
            result_str += f"│    └── [{', '.join(map(str, zstate))}] => {count}\n"

        result_str += f"├── {len(self.fidelities)} executions\n"
        result_str += f"├── {len(self.amplitudes)} amplitudes\n"
        result_str += f"└── {len(self.cstates)} samples"

        return result_str

    def _repr_html_(self):
        html_output = "<table><tbody>"

        # QCSResults Header
        html_output += '<tr><td colspan=2 style="text-align:center;"><strong>QCSResults</strong></td></tr>'
        html_output += "<tr><td colspan=2><hr></td></tr>"

        # Simulator Information
        html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Simulator</strong></td></tr>'
        html_output += f'<tr><td colspan=2 style="text-align:center;">{self.simulator} {self.version}</td></tr>'
        html_output += "<tr><td colspan=2><hr></td></tr>"

        # Timings
        html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Timings</strong></td></tr>'
        for key, value in self.timings.items():
            html_output += f'<tr><td style="text-align:left;">{key} time</td><td>{value}s</td></tr>'
        html_output += "<tr><td colspan=2><hr></td></tr>"

        # Fidelity Estimates
        if self.fidelities:
            html_output += "<tr><td colspan=2></td></tr>"
            html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Fidelity estimate</strong></td></tr>'
            if len(self.fidelities) == 1:
                html_output += f'<tr><td style="text-align:left;">Single run value</td><td>{round(self.fidelities[0], 3)}</td></tr>'
            else:
                html_output += f'<tr><td style="text-align:left;">Mean</td><td>{round(mean(self.fidelities), 3)}</td></tr>'
                html_output += f'<tr><td style="text-align:left;">Median</td><td>{round(median(self.fidelities), 3)}</td></tr>'
                html_output += f'<tr><td style="text-align:left;">Standard Deviation</td><td>{round(stdev(self.fidelities), 3)}</td></tr>'
            html_output += "<tr><td colspan=2><hr></td></tr>"

        # Average Multiqubit Error Estimates
        if self.avggateerrors:
            html_output += "<tr><td colspan=2></td></tr>"
            html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Average multiqubit error estimate</strong></td></tr>'
            if len(self.avggateerrors) == 1:
                html_output += f'<tr><td style="text-align:left;">Single run value</td><td>{round(self.avggateerrors[0], 3)}</td></tr>'
            else:
                html_output += f'<tr><td style="text-align:left;">Mean</td><td>{round(mean(self.avggateerrors), 3)}</td></tr>'
                html_output += f'<tr><td style="text-align:left;">Median</td><td>{round(median(self.avggateerrors), 3)}</td></tr>'
                html_output += f'<tr><td style="text-align:left;">Standard Deviation</td><td>{round(stdev(self.avggateerrors), 3)}</td></tr>'
            html_output += "<tr><td colspan=2><hr></td></tr>"

        # Statistics
        html_output += "<tr><td colspan=2></td></tr>"
        html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Statistics</strong></td></tr>'
        html_output += f'<tr><td style="text-align:left;">Number of executions</td><td>{len(self.fidelities)}</td></tr>'
        html_output += f'<tr><td style="text-align:left;">Number of samples</td><td>{len(self.cstates)}</td></tr>'
        html_output += f'<tr><td style="text-align:left;">Number of amplitudes</td><td>{len(self.amplitudes)}</td></tr>'
        html_output += "<tr><td colspan=2><hr></td></tr>"

        # Sampled Classical States
        if self.cstates:
            html_output += "<tr><td colspan=2></td></tr>"
            html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Samples</strong></td></tr>'
            hist = self.histogram()
            outcomes = sorted(hist, key=hist.get, reverse=True)[0:10]
            for bs in outcomes:
                html_output += f'<tr><td style="text-align:left;font-family: monospace;">{hex(bitvec_to_int(bs))}</td><td style="text-align:left;font-family: monospace;">{bs.to01()}</td><td style="font-family: monospace;">{hist[bs]}</td></tr>'
            html_output += "<tr><td colspan=2><hr></td></tr>"

        # Statevector Amplitudes
        if self.amplitudes:
            html_output += "<tr><td colspan=2></td></tr>"
            html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Statevector Amplitudes</strong></td></tr>'
            for bs, amp in self.amplitudes.items():
                html_output += f'<tr><td style="text-align:left;">{hex(bitvec_to_int(bs.bits))}</td><td style="text-align:left;">{bs.to01()}</td><td>{amp:.3f}</td></tr>'
            html_output += "<tr><td colspan=2><hr></td></tr>"

        if len(self.zstates) > 0 and any(len(zstate) > 0 for zstate in self.zstates):
            html_output += "<tr><td colspan=2></td></tr>"
            html_output += '<tr><td colspan=2 style="text-align:center;"><strong>Z States</strong></td></tr>'
            z_hist = self.histzvars()
            for zstate, count in z_hist.items():
                html_output += f'<tr><td style="text-align:left;font-family: monospace;">[{", ".join(map(str, zstate))}]</td><td style="text-align:left;font-family: monospace;">{count}</td></tr>'
            html_output += "<tr><td colspan=2><hr></td></tr>"

        html_output += "</tbody></table>"

        return html_output

    def histogram(self):
        """Histogram of the obtained classical states' occurrences.

        Returns:
            A dictionary of classical states (bitarray) and their occurrences
            (float).

        Raises:
            TypeError: If a non QCSResults object is passed.
        """
        hist = {}

        for cstate in self.cstates:
            if cstate in hist:
                hist[cstate] += 1
            else:
                hist[cstate] = 1

        return hist

    def histzvars(self):
        """
        Histogram of the obtained zstates' occurrences.
        """
        hist = {}

        for zstate in self.zstates:
            i = tuple(zstate)  # Make it hashable
            if i in hist:
                hist[i] += 1
            else:
                hist[i] = 1

        return hist

    def saveproto(self, file):
        """Save QCSResults object to a Protocol Buffers file.

        Examples:

            >>> from mimiqcircuits import *
            >>> from symengine import *
            >>> import tempfile
            >>> x, y = symbols("x y")
            >>> c = Circuit()
            >>> c.push(GateH(), 0)
            1-qubit circuit with 1 instructions:
            └── H @ q[0]
            <BLANKLINE>
            >>> conn = MimiqConnection()
            >>> conn.connect()
            Connection:
            ├── url: https://mimiq.qperfect.io/api
            ├── Computing time: 597/10000 minutes
            ├── Executions: 451/10000
            ├── Max time limit per request: 180 minutes
            └── status: open
            <BLANKLINE>
            >>> job = conn.execute(c)
            >>> res = conn.get_result(job)
            >>> res
            QCSResults:
            ├── simulator: MIMIQ-StateVector 0.18.0
            ├── timings:
            │    ├── parse time: 5.9552e-05s
            │    ├── apply time: 1.6682e-05s
            │    ├── total time: 0.00019081399999999998s
            │    ├── compression time: 4.575e-06s
            │    └── sample time: 5.299e-05s
            ├── fidelity estimate: 1
            ├── average multi-qubit gate error estimate: 0
            ├── most sampled:
            │    ├── bs"1" => 513
            │    └── bs"0" => 487
            ├── 1 executions
            ├── 0 amplitudes
            └── 1000 samples

            >>> tmpfile = tempfile.NamedTemporaryFile(suffix=".pb", delete=True)
            >>> res.saveproto(tmpfile.name)
            7169
            >>> res.loadproto(tmpfile.name)
            QCSResults:
            ├── simulator: MIMIQ-StateVector 0.18.0
            ├── timings:
            │    ├── parse time: 5.9552e-05s
            │    ├── apply time: 1.6682e-05s
            │    ├── total time: 0.00019081399999999998s
            │    ├── compression time: 4.575e-06s
            │    └── sample time: 5.299e-05s
            ├── fidelity estimate: 1
            ├── average multi-qubit gate error estimate: 0
            ├── most sampled:
            │    ├── bs"1" => 513
            │    └── bs"0" => 487
            ├── 1 executions
            ├── 0 amplitudes
            └── 1000 samples

            Note:
                This example uses a temporary file to demonstrate the save and load functionality.
                You can save your file with any name at any location using:

                .. code-block:: python

                    res.saveproto("example.pb")
                    res.loadproto("example.pb")
        """
        if isinstance(file, str):
            with open(file, "wb") as f:
                return f.write(toproto_qcsr(self).SerializeToString())
        elif hasattr(file, "write"):
            return file.write(toproto_qcsr(self).SerializeToString())

    @staticmethod
    def loadproto(file):
        """Load QCSResults object from a Protocol Buffers file.

        The :func:`loadproto` method is a static method and should be called on the class,
        not on an instance of the class.

        Note:

            Look for example in :func:`QCSResults.saveproto`
        """
        from mimiqcircuits.proto import qcsresults_pb2

        qcs_results_proto = qcsresults_pb2.QCSResults()

        if isinstance(file, str):
            with open(file, "rb") as f:
                qcs_results_proto.ParseFromString(f.read())
        elif hasattr(file, "read"):
            qcs_results_proto.ParseFromString(file.read())

        return fromproto_qcsr(qcs_results_proto)


__all__ = ["QCSResults"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqcircuits.operations.power import Power
from mimiqcircuits.operations.inverse import Inverse
from mimiqcircuits.operations.gates.standard.s import GateS
from mimiqcircuits.operations.gates.standard.u import GateU
from symengine import pi
import mimiqcircuits as mc


class GateT(Power):
    r""" Single qubit T gate.

    **Matrix representation:**

    .. math::
        \operatorname{T} = \begin{pmatrix}
            1 & 0 \\
            0 & \exp\left(\frac{i\pi}{4}\right)
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateT()
        T
        >>> GateT().matrix()
        [1.0, 0]
        [0, 0.707106781186548 + 0.707106781186548*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateT(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── T @ q[0]
        <BLANKLINE>
        >>> GateT().power(2), GateT().inverse()
        (S, T†)
        >>> GateT().decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, (1/4)*pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    _name = "T"

    def __init__(self):
        super().__init__(GateS(), 1 / 2)

    def isopalias(self):
        return True

    def inverse(self):
        return GateTDG()

    def _power(self, p):
        # T^2 * T^2 * T^2 * T^2 = S^2 * S^2 = Z * Z = ID
        if p % 8 == 0:
            return mc.GateID()

        # T^(8n + 1) = T
        if p % 8 == 1:
            return GateT()

        # T^(8n - 1) = T†
        if p % 8 == 7:
            return GateTDG()

        # T^(4n) = Z^n
        if p % 4 == 0:
            return mc.GateZ().power(p / 4)

        # T^(2n) = S^n
        if p % 2 == 0:
            return mc.GateS().power(p / 2)

        return mc.Power(self, p)

    def __str__(self):
        return f"{self.name}"

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(0, 0, pi / 4), q)
        return circ


class GateTDG(Inverse):
    r"""Single qubit T-dagger gate (conjugate transpose of the T gate).

    **Matrix representation:**

    .. math::
        \operatorname{T}^\dagger = \begin{pmatrix}
            1 & 0 \\
            0 & \exp\left(\frac{-i\pi}{4}\right)
        \end{pmatrix}

    Examples:
        >>> from mimiqcircuits import *
        >>> GateTDG()
        T†
        >>> GateTDG().matrix()
        [1.0, 0]
        [0, 0.707106781186547 - 0.707106781186547*I]
        <BLANKLINE>
        >>> c = Circuit().push(GateTDG(), 0)
        >>> c
        1-qubit circuit with 1 instructions:
        └── T† @ q[0]
        <BLANKLINE>
        >>> GateTDG().power(2), GateTDG().inverse()
        (T†**2, T)
        >>> GateTDG().decompose()
        1-qubit circuit with 1 instructions:
        └── U(0, 0, (-1/4)*pi, 0.0) @ q[0]
        <BLANKLINE>
    """

    def __init__(self):
        super().__init__(GateT())

    def isopalias(self):
        return True

    def inverse(self):
        return GateT()

    def _decompose(self, circ, qubits, bits, zvars):
        q = qubits[0]
        circ.push(GateU(0, 0, -pi / 4), q)
        return circ



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
from symengine import pi


class Diffusion(mc.Gate):
    """
    Grover's diffusion operator.

    Args:
        num_qubits (int): The number of qubits.

    Raises:
        ValueError: If the number of qubits is not an integer or less than 1.

    Returns:
        Diffusion: Grover's diffusion operator.

    Attributes:
        num_qubits (int): The number of qubits for the diffusion operator.

    Examples:

        >>> from mimiqcircuits import *
        >>> c = Circuit()
        >>> c.push(Diffusion(2), 1, 2)
        3-qubit circuit with 1 instructions:
        └── Diffusion @ q[1,2]
        <BLANKLINE>

    """

    _name = "Diffusion"
    _num_qregs = 1
    _num_qubits = None

    def __init__(self, num_qubits):
        if not isinstance(num_qubits, int):
            raise ValueError("Number of qubits must be an integer.")

        if num_qubits < 1:
            raise ValueError("Number of qubits must be at least 1.")

        super().__init__()

        self._num_qubits = num_qubits
        self._qregsizes = [num_qubits]
        self._params = [num_qubits]
        self._params = [num_qubits]

    def __new__(cls, *args):
        if len(args) == 0:
            return mc.LazyExpr(Diffusion, mc.LazyArg())
        elif len(args) == 1:
            return object.__new__(cls)
        else:
            raise ValueError("Invalid number of arguments.")

    def __str__(self):
        return "Diffusion"

    def _matrix(self):
        raise NotImplementedError(
            "Matrix representation for Diffusion operator is not implemented."
        )

    def _decompose(self, circ, qubits, bits, zvars):
        circ.push(mc.GateRY(pi / 2), qubits)
        circ.push(mc.control(self.num_qubits - 1, mc.GateZ()), *qubits)
        circ.push(mc.GateRY(-pi / 2), qubits)
        return circ


__all__ = ["Diffusion"]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2032-2024 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import mimiqcircuits as mc
from mimiqcircuits.matrices import cis
from symengine import I, Matrix, cos, sin, pi


class GateRXX(mc.Gate):
    r"""Two qubit RXX gate (rotation about XX).

    This gate is symmetric, and is maximally entangling at :math:`(\theta = \frac{\pi}{2})`

    **Matrix representation:**

    .. math::
        \operatorname{RXX}(\theta) =\begin{pmatrix}
            \cos(\frac{\theta}{2}) & 0 & 0 & -i\sin(\frac{\theta}{2}) \\
            0 & \cos(\frac{\theta}{2}) & -i\sin(\frac{\theta}{2}) & 0 \\
            0 & -i\sin(\frac{\theta}{2}) & \cos(\frac{\theta}{2}) & 0 \\
            -i\sin(\frac{\theta}{2}) & 0 & 0 & \cos(\frac{\theta}{2})
        \end{pmatrix}

    Parameters:
        theta: The angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta = Symbol('theta')
        >>> GateRXX(theta)
        RXX(theta)
        >>> c = Circuit()
        >>> c.push(GateRXX(theta), 0, 1)
        2-qubit circuit with 1 instructions:
        └── RXX(theta) @ q[0,1]
        <BLANKLINE>
        >>> GateRXX(theta).power(2), GateRXX(theta).inverse()
        (RXX(theta)**2, RXX(-theta))
        >>> GateRXX(theta).matrix()
        [cos((1/2)*theta), 0, 0, -I*sin((1/2)*theta)]
        [0, cos((1/2)*theta), -I*sin((1/2)*theta), 0]
        [0, -I*sin((1/2)*theta), cos((1/2)*theta), 0]
        [-I*sin((1/2)*theta), 0, 0, cos((1/2)*theta)]
        <BLANKLINE>
        >>> GateRXX(theta).decompose()
        2-qubit circuit with 7 instructions:
        ├── H @ q[0]
        ├── H @ q[1]
        ├── CX @ q[0], q[1]
        ├── RZ(theta) @ q[1]
        ├── CX @ q[0], q[1]
        ├── H @ q[1]
        └── H @ q[0]
        <BLANKLINE>
    """

    _name = "RXX"

    _num_qubits = 2
    _qregsizes = [2]

    _parnames = ("theta",)

    def __init__(self, theta):
        self.theta = theta

    def _matrix(self):
        theta = self.theta
        cos2 = cos(theta / 2)
        sin2 = sin(theta / 2)

        return Matrix(
            [
                [cos2, 0, 0, -I * sin2],
                [0, cos2, -I * sin2, 0],
                [0, -I * sin2, cos2, 0],
                [-I * sin2, 0, 0, cos2],
            ]
        )

    def inverse(self):
        return GateRXX(-self.theta)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(mc.GateH(), a)
        circ.push(mc.GateH(), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateRZ(self.theta), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateH(), b)
        circ.push(mc.GateH(), a)
        return circ


class GateRYY(mc.Gate):
    r"""Two qubit RYY gate (rotation about YY).

    This gate is symmetric, and is maximally entangling at :math:`(\theta = \frac{\pi}{2})`

    **Matrix representation:**

    .. math::
        \operatorname{RYY}(\theta) =\begin{pmatrix}
            \cos(\frac{\theta}{2}) & 0 & 0 & i\sin(\frac{\theta}{2}) \\
            0 & \cos(\frac{\theta}{2}) & -i\sin(\frac{\theta}{2}) & 0 \\
            0 & -i\sin(\frac{\theta}{2}) & \cos(\frac{\theta}{2}) & 0 \\
            i\sin(\frac{\theta}{2}) & 0 & 0 & \cos(\frac{\theta}{2})
        \end{pmatrix}

    Parameters:
        theta (float): The angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta = Symbol('theta')
        >>> GateRYY(theta)
        RYY(theta)
        >>> GateRYY(theta).matrix()
        [cos((1/2)*theta), 0, 0, I*sin((1/2)*theta)]
        [0, cos((1/2)*theta), -I*sin((1/2)*theta), 0]
        [0, -I*sin((1/2)*theta), cos((1/2)*theta), 0]
        [I*sin((1/2)*theta), 0, 0, cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateRYY(theta), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── RYY(theta) @ q[0,1]
        <BLANKLINE>
        >>> GateRYY(theta).power(2), GateRYY(theta).inverse()
        (RYY(theta)**2, RYY(-theta))
        >>> GateRYY(theta).decompose()
        2-qubit circuit with 7 instructions:
        ├── RX((1/2)*pi) @ q[0]
        ├── RX((1/2)*pi) @ q[1]
        ├── CX @ q[0], q[1]
        ├── RZ(theta) @ q[1]
        ├── CX @ q[0], q[1]
        ├── RX((-1/2)*pi) @ q[0]
        └── RX((-1/2)*pi) @ q[1]
        <BLANKLINE>
    """

    _name = "RYY"

    _num_qubits = 2
    _qregsizes = [2]

    _parnames = ("theta",)

    def __init__(self, theta):
        self.theta = theta

    def _matrix(self):
        theta = self.theta
        cos2 = cos(theta / 2)
        sin2 = sin(theta / 2)

        return Matrix(
            [
                [cos2, 0, 0, I * sin2],
                [0, cos2, -I * sin2, 0],
                [0, -I * sin2, cos2, 0],
                [I * sin2, 0, 0, cos2],
            ]
        )

    def inverse(self):
        return GateRYY(-self.theta)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(mc.GateRX(pi / 2), a)
        circ.push(mc.GateRX(pi / 2), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateRZ(self.theta), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateRX(-pi / 2), a)
        circ.push(mc.GateRX(-pi / 2), b)

        return circ


class GateRZZ(mc.Gate):
    r"""Two qubit RZZ gate (rotation about ZZ)..

    This gate is symmetric, and is maximally entangling at :math:`(\theta = \frac{\pi}{2})`

    **Matrix representation:**

    .. math::
        \operatorname{RZZ}(\theta) = \begin{pmatrix}
            e^{-i\frac{\theta}{2}} & 0 & 0 & 0 \\
            0 & e^{i\frac{\theta}{2}} & 0 & 0 \\
            0 & 0 & e^{i\frac{\theta}{2}} & 0 \\
            0 & 0 & 0 & e^{-i\frac{\theta}{2}}
        \end{pmatrix}

    Parameters:
        theta (float): The angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta = Symbol('theta')
        >>> GateRZZ(theta)
        RZZ(theta)
        >>> GateRZZ(theta).matrix()
        [exp(-1/2*I*theta), 0, 0, 0]
        [0, exp(1/2*I*theta), 0, 0]
        [0, 0, exp(1/2*I*theta), 0]
        [0, 0, 0, exp(-1/2*I*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateRZZ(theta), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── RZZ(theta) @ q[0,1]
        <BLANKLINE>
        >>> GateRZZ(theta).power(2), GateRZZ(theta).inverse()
        (RZZ(theta)**2, RZZ(-theta))
        >>> GateRZZ(theta).decompose()
        2-qubit circuit with 3 instructions:
        ├── CX @ q[0], q[1]
        ├── RZ(theta) @ q[1]
        └── CX @ q[0], q[1]
        <BLANKLINE>
    """

    _num_qubits = 2
    _name = "RZZ"
    _parnames = ("theta",)
    _qregsizes = [2]

    def __init__(self, theta):
        self.theta = theta

    def _matrix(self):
        return Matrix(
            [
                [cis(-self.theta / 2), 0, 0, 0],
                [0, cis(self.theta / 2), 0, 0],
                [0, 0, cis(self.theta / 2), 0],
                [0, 0, 0, cis(-self.theta / 2)],
            ]
        )

    def inverse(self):
        return GateRZZ(-self.theta)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateRZ(self.theta), b)
        circ.push(mc.GateCX(), a, b)
        return circ


class GateRZX(mc.Gate):
    r"""Two qubit RZX gate.

    This gate is maximally entangling at :math:`(\theta = \frac{\pi}{2})`

    **Matrix representation:**

    .. math::
        \operatorname{RZX}(\theta) =\begin{pmatrix}
            \cos(\frac{\theta}{2}) & -i\sin(\frac{\theta}{2}) & 0 & 0 \\
            -i\sin(\frac{\theta}{2}) & \cos(\frac{\theta}{2}) & 0 & 0 \\
            0 & 0 & \cos(\frac{\theta}{2}) & i\sin(\frac{\theta}{2}) \\
            0 & 0 & i\sin(\frac{\theta}{2}) & \cos(\frac{\theta}{2})
        \end{pmatrix}

    Parameters:
        theta (float): The angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta = Symbol('theta')
        >>> GateRZX(theta)
        RZX(theta)
        >>> GateRZX(theta).matrix()
        [cos((1/2)*theta), -I*sin((1/2)*theta), 0, 0]
        [-I*sin((1/2)*theta), cos((1/2)*theta), 0, 0]
        [0, 0, cos((1/2)*theta), I*sin((1/2)*theta)]
        [0, 0, I*sin((1/2)*theta), cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateRZX(theta), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── RZX(theta) @ q[0,1]
        <BLANKLINE>
        >>> GateRZX(theta).power(2), GateRZX(theta).inverse()
        (RZX(theta)**2, RZX(-theta))
        >>> GateRZX(theta).decompose()
        2-qubit circuit with 5 instructions:
        ├── H @ q[1]
        ├── CX @ q[0], q[1]
        ├── RZ(theta) @ q[1]
        ├── CX @ q[0], q[1]
        └── H @ q[1]
        <BLANKLINE>
    """

    _num_qubits = 2
    _name = "RZX"
    _parnames = ("theta",)
    _qregsizes = [2]

    def __init__(self, theta):
        self.theta = theta

    def _matrix(self):
        theta = self.theta
        cos2 = cos(theta / 2)
        sin2 = sin(theta / 2)

        return Matrix(
            [
                [cos2, -I * sin2, 0, 0],
                [-I * sin2, cos2, 0, 0],
                [0, 0, cos2, I * sin2],
                [0, 0, I * sin2, cos2],
            ]
        )

    def inverse(self):
        return GateRZX(-self.theta)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits
        circ.push(mc.GateH(), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateRZ(self.theta), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateH(), b)

        return circ


class GateXXplusYY(mc.Gate):
    r"""Two qubit parametric XXplusYY gate.

    Also known as an XY gate. Its action is to induce a coherent rotation by some angle between :math:`\ket{10}` and :math:`\ket{01}`.

    **Matrix representation:**

    .. math::
        \operatorname{(XX+YY)}(\theta, \beta)= \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & \cos(\frac{\theta}{2}) & -i\sin(\frac{\theta}{2})e^{i\beta} & 0\\
            0 & -i\sin(\frac{\theta}{2})e^{-i\beta} & \cos(\frac{\theta}{2}) & 0\\
            0 & 0 & 0 & 1
        \end{pmatrix}

    Parameters:
        theta: The angle in radians.
        beta: The phase angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta, beta = symbols('theta beta')
        >>> GateXXplusYY(theta, beta)
        XXplusYY(theta, beta)
        >>> GateXXplusYY(theta, beta).matrix()
        [1.0, 0, 0, 0]
        [0, cos((1/2)*theta), (sin(beta) - I*cos(beta))*sin((1/2)*theta), 0]
        [0, I*(I*sin(beta) - cos(beta))*sin((1/2)*theta), cos((1/2)*theta), 0]
        [0, 0, 0, 1.0]
        <BLANKLINE>
        >>> c = Circuit().push(GateXXplusYY(theta, beta), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── XXplusYY(theta, beta) @ q[0,1]
        <BLANKLINE>
        >>> GateXXplusYY(theta, beta).power(2), GateXXplusYY(theta, beta).inverse()
        (XXplusYY(theta, beta)**2, XXplusYY(-theta, beta))
        >>> GateXXplusYY(theta, beta).decompose()
        2-qubit circuit with 14 instructions:
        ├── RZ(beta) @ q[0]
        ├── RZ((-1/2)*pi) @ q[1]
        ├── SX @ q[1]
        ├── RZ((1/2)*pi) @ q[1]
        ├── S @ q[0]
        ├── CX @ q[1], q[0]
        ├── RY((-1/2)*theta) @ q[1]
        ├── RY((-1/2)*theta) @ q[0]
        ├── CX @ q[1], q[0]
        ├── S† @ q[0]
        ├── RZ((-1/2)*pi) @ q[1]
        ├── SX† @ q[1]
        ├── RZ((1/2)*pi) @ q[1]
        └── RZ(-beta) @ q[0]
        <BLANKLINE>
    """

    _num_qubits = 2
    _name = "XXplusYY"
    _parnames = ("theta", "beta")
    _qregsizes = [2]

    def __init__(self, theta, beta):
        self.theta = theta
        self.beta = beta

    def _matrix(self):
        theta = self.theta
        beta = self.beta
        cos2 = cos(theta / 2)
        sin2 = sin(theta / 2)

        return Matrix(
            [
                [1, 0, 0, 0],
                [0, cos2, -I * sin2 * cis(beta), 0],
                [0, -I * sin2 * cis(-beta), cos2, 0],
                [0, 0, 0, 1],
            ]
        )

    def inverse(self):
        return GateXXplusYY(-self.theta, self.beta)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(mc.GateRZ(self.beta), a)
        circ.push(mc.GateRZ(-pi / 2), b)
        circ.push(mc.GateSX(), b)
        circ.push(mc.GateRZ(pi / 2), b)
        circ.push(mc.GateS(), a)
        circ.push(mc.GateCX(), b, a)
        circ.push(mc.GateRY(-self.theta / 2), b)
        circ.push(mc.GateRY(-self.theta / 2), a)
        circ.push(mc.GateCX(), b, a)
        circ.push(mc.GateSDG(), a)
        circ.push(mc.GateRZ(-pi / 2), b)
        circ.push(mc.GateSXDG(), b)
        circ.push(mc.GateRZ(pi / 2), b)
        circ.push(mc.GateRZ(-self.beta), a)

        return circ


class GateXXminusYY(mc.Gate):
    r"""Two qubit parametric GateXXminusYY gate.

    Its action is to induce a coherent rotation by some angle between :math:`\ket{00}` and :math:`\ket{11}`

    **Matrix representation:**

    .. math::
        \operatorname{(XX-YY)}(\theta, \beta)=\begin{pmatrix}
            \cos(\frac{\theta}{2}) & 0 & 0 & -i\sin(\frac{\theta}{2})e^{-i\beta} \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 1 & 0 \\
            -i\sin(\frac{\theta}{2})e^{i\beta} & 0 & 0 & \cos(\frac{\theta}{2})
        \end{pmatrix}

    Parameters:
        theta (float): The angle in radians.
        beta (float): The angle in radians.

    Examples:
        >>> from mimiqcircuits import *
        >>> from symengine import *
        >>> theta, beta = symbols('theta beta')
        >>> GateXXminusYY(theta, beta)
        XXminusYY(theta, beta)
        >>> GateXXminusYY(theta, beta).matrix()
        [cos((1/2)*theta), 0, 0, I*(I*sin(beta) - cos(beta))*sin((1/2)*theta)]
        [0, 1.0, 0, 0]
        [0, 0, 1.0, 0]
        [(sin(beta) - I*cos(beta))*sin((1/2)*theta), 0, 0, cos((1/2)*theta)]
        <BLANKLINE>
        >>> c = Circuit().push(GateXXminusYY(theta, beta), 0, 1)
        >>> c
        2-qubit circuit with 1 instructions:
        └── XXminusYY(theta, beta) @ q[0,1]
        <BLANKLINE>
        >>> GateXXminusYY(theta, beta).power(2), GateXXminusYY(theta, beta).inverse()
        (XXminusYY(theta, beta)**2, XXminusYY(-theta, beta))
        >>> GateXXminusYY(theta, beta).decompose()
        2-qubit circuit with 14 instructions:
        ├── RZ(-beta) @ q[1]
        ├── RZ((-1/2)*pi) @ q[0]
        ├── SX @ q[0]
        ├── RZ((1/2)*pi) @ q[0]
        ├── S @ q[1]
        ├── CX @ q[0], q[1]
        ├── RY((1/2)*theta) @ q[0]
        ├── RY((-1/2)*theta) @ q[1]
        ├── CX @ q[0], q[1]
        ├── S† @ q[1]
        ├── RZ((-1/2)*pi) @ q[0]
        ├── SX† @ q[0]
        ├── RZ((1/2)*pi) @ q[0]
        └── RZ(beta) @ q[1]
        <BLANKLINE>
    """

    _num_qubits = 2
    _name = "XXminusYY"
    _parnames = ("theta", "beta")
    _qregsizes = [2]

    def __init__(self, theta, beta):
        self.theta = theta
        self.beta = beta

    def _matrix(self):
        theta = self.theta
        beta = self.beta
        cos2 = cos(theta / 2)
        sin2 = sin(theta / 2)

        return Matrix(
            [
                [cos2, 0, 0, -I * sin2 * cis(-beta)],
                [0, 1, 0, 0],
                [0, 0, 1, 0],
                [-I * sin2 * cis(beta), 0, 0, cos2],
            ]
        )

    def inverse(self):
        return GateXXminusYY(-self.theta, self.beta)

    def _decompose(self, circ, qubits, bits, zvars):
        a, b = qubits

        circ.push(mc.GateRZ(-self.beta), b)
        circ.push(mc.GateRZ(-pi / 2), a)
        circ.push(mc.GateSX(), a)
        circ.push(mc.GateRZ(pi / 2), a)
        circ.push(mc.GateS(), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateRY(self.theta / 2), a)
        circ.push(mc.GateRY(-self.theta / 2), b)
        circ.push(mc.GateCX(), a, b)
        circ.push(mc.GateSDG(), b)
        circ.push(mc.GateRZ(-pi / 2), a)
        circ.push(mc.GateSXDG(), a)
        circ.push(mc.GateRZ(pi / 2), a)
        circ.push(mc.GateRZ(self.beta), b)

        return circ
